# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import torch
import time
from multiprocessing import shared_memory
import numpy as np
import torch.distributed as dist
import toml
from transformers import AutoConfig
import cosmos_rl.utils.util as util
from transformers import AutoTokenizer
from cosmos_rl.policy.model import ModelRegistry, WeightMapper
import msgpack
import threading
from cosmos_rl.policy.trainer.grpo_trainer import GRPOTrainer
from cosmos_rl.policy.trainer import Trainer
from cosmos_rl.policy.trainer.sft_trainer import SFTTrainer
from cosmos_rl.rollout.vllm_rollout.vllm_rollout_worker import vLLMRolloutWorker
from cosmos_rl.rollout.vllm_rollout.vllm_rollout import vLLMRollout
from cosmos_rl.rollout import State
import types
from cosmos_rl.dispatcher.command import (
    PolicyToRolloutUnicastCommand,
    PolicyToPolicyUnicastCommand,
    PolicyToPolicyBroadcastCommand,
)
from cosmos_rl.utils.parallelism_map import (
    ParallelTopoMapperGroup,
    ParallelTopoMapper,
    ParallelizedShardMapper,
    WeightSyncInstructionsGroup,
)
from cosmos_rl.utils.parallelism import ParallelismConfig, ParallelDims
from cosmos_rl.utils.distributed import (
    init_distributed,
    destroy_distributed,
)
from cosmos_rl.dispatcher.protocol import Role
from cosmos_rl.policy.model.gpt.weight_converter import convert_weight_from_hf
from cosmos_rl.policy.model.gpt.weight_mapper import GPTWeightMapper
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.comm.base import CommMixin
from cosmos_rl.utils.distributed import HighAvailabilitylNccl
from cosmos_rl.utils.pynccl import (
    create_nccl_uid,
    create_nccl_comm,
    nccl_send,
    nccl_recv,
    nccl_broadcast,
)
import cosmos_rl.utils.distributed as dist_util
from cosmos_rl.utils.logging import logger
import asyncio
from cosmos_rl.dispatcher.data.packer import (
    DecoderOnlyLLMDataPacker,
)
import cosmos_rl.utils.distributed as dist_utils
import uuid
from cosmos_rl.utils.ulysses import (
    slice_inputs_for_ulysses,
)
from cosmos_rl.utils.sequence_packing import (
    pack_sequences_info_collect,
    pack_sequences_for_masks,
    pack_sequences_for_labels,
)
from torch.utils.data import DataLoader, DistributedSampler, Sampler
from cosmos_rl.policy.trainer.sft_trainer import collate_fn, construct_dataset


POLICY_WORLD_SIZE = 4
ROLLOUT_WORLD_SIZE = 4


class TestModel:
    model_type = "qwen2"
    model_path = "Qwen/Qwen2.5-3B-Instruct"
    num_hidden_layers = 16

    def __init__(self, device, parallel_dims, freeze_params: bool = False):
        self.sorted_hf_key_n_rank = [
            ("model.layers.9.input_layernorm.weight", torch.Size([1024])),
            ("model.layers.9.mlp.down_proj.weight", torch.Size([1024, 11008])),
            ("model.layers.9.mlp.gate_proj.weight", torch.Size([5504, 2048])),
            ("model.layers.9.mlp.up_proj.weight", torch.Size([5504, 2048])),
            ("model.layers.9.post_attention_layernorm.weight", torch.Size([1024])),
            ("model.layers.9.self_attn.k_proj.bias", torch.Size([128])),
            ("model.layers.9.self_attn.k_proj.weight", torch.Size([128, 2048])),
            ("model.layers.9.self_attn.o_proj.weight", torch.Size([1024, 2048])),
            ("model.layers.9.self_attn.q_proj.bias", torch.Size([1024])),
            ("model.layers.9.self_attn.q_proj.weight", torch.Size([1024, 2048])),
            ("model.layers.9.self_attn.v_proj.bias", torch.Size([128])),
            ("model.layers.9.self_attn.v_proj.weight", torch.Size([128, 2048])),
            ("lm_head.weight", torch.Size([75968, 2048])),
            ("model.norm.weight", torch.Size([1024])),
            ("model.embed_tokens.weight", torch.Size([75968, 2048])),
        ]

        self.parallel_spec = [
            ("model.layers.9.input_layernorm.weight", {}),
            ("model.layers.9.mlp.down_proj.weight", {"tp": 1}),
            ("model.layers.9.mlp.gate_proj.weight", {"tp": 0}),
            ("model.layers.9.mlp.up_proj.weight", {"tp": 0}),
            ("model.layers.9.post_attention_layernorm.weight", {}),
            ("model.layers.9.self_attn.k_proj.bias", {"tp": 0}),
            ("model.layers.9.self_attn.k_proj.weight", {"tp": 0}),
            ("model.layers.9.self_attn.o_proj.weight", {"tp": 0}),
            ("model.layers.9.self_attn.q_proj.bias", {"tp": 0}),
            ("model.layers.9.self_attn.q_proj.weight", {"tp": 0}),
            ("model.layers.9.self_attn.v_proj.bias", {"tp": 0}),
            ("model.layers.9.self_attn.v_proj.weight", {"tp": 0}),
            ("lm_head.weight", {"tp": 0}),
            ("model.norm.weight", {}),
            ("model.embed_tokens.weight", {"tp": 0}),
        ]

        self.keys_to_freeze = (
            [
                "model.layers.9.mlp.down_proj.weight",
                "model.layers.9.mlp.gate_proj.weight",
                "model.layers.9.mlp.up_proj.weight",
            ]
            if freeze_params
            else []
        )

        self.sorted_hf_key_n_rank.sort(key=lambda x: x[0])

        self.config = AutoConfig.from_pretrained(self.model_path)
        self.device = device
        self.parallel_dims = parallel_dims
        self.tensors = [
            (
                k,
                (
                    torch.arange(v.numel(), dtype=torch.float32, device=self.device)
                    .reshape(v)
                    .to(self.device)
                    * 0.001
                ).requires_grad_(True)
                if k not in self.keys_to_freeze
                else (
                    torch.arange(v.numel(), dtype=torch.float32, device=self.device)
                    .reshape(v)
                    .to(self.device)
                    * 0.001
                ).requires_grad_(False),
            )
            for k, v in self.sorted_hf_key_n_rank
        ]
        self.sharded_tensors = {}
        for k, v in self.tensors:
            self.sharded_tensors[k] = convert_weight_from_hf(
                v, k, self.model_type, self.parallel_dims
            )[1]
        self.sorted_sharded_params = [
            (k, self.sharded_tensors[k].ndim) for k, _ in self.sorted_hf_key_n_rank
        ]
        self.weight_mapper = GPTWeightMapper(self.config)

    def get_trainable_params(self):
        trainable_params = set()
        for k, v in self.sharded_tensors.items():
            if v.requires_grad:
                trainable_params.add(k)
        return trainable_params


class TestPolicy:
    def __init__(
        self, name, policy_world_size, rollouts_comm, freeze_params: bool = False
    ):
        self.local_rank = int(os.environ.get("LOCAL_RANK", 0))
        self.device = torch.device(f"cuda:{self.local_rank}")
        self.global_rank = int(os.environ.get("RANK", 0))
        self.role = Role.POLICY
        self.world_size = policy_world_size
        policy_parallelism_dims = ParallelismConfig(
            dp_shard_size=2, cp_size=1, tp_size=2, pp_size=1
        )
        self.parallel_dims = ParallelDims.from_config(
            policy_parallelism_dims,
        )
        self.parallel_dims.build_mesh(device_type="cuda")
        self.model = TestModel(self.device, self.parallel_dims, freeze_params)
        self.parallel_mapper = ParallelTopoMapperGroup(
            self.parallel_dims,
            self.model.config,
            True,
            self.model,
            self.model.weight_mapper,
        )
        self.replica_name = name
        self.rollouts_comm = rollouts_comm
        self.policy_to_rollout_insts = None
        self.map_w_from_policy_to_rollout = self.model.sharded_tensors
        self.model.sorted_hf_key_n_rank = self.model.sorted_sharded_params
        self.p2r_nccl_uuids = rollouts_comm
        self.train_stream = torch.cuda.Stream()
        self.config = CosmosConfig()
        self.config.train.param_dtype = "float32"

        self.prepare_trainable_params()

    def execute_policy_to_rollout_unicast(self, command: PolicyToRolloutUnicastCommand):
        pass

    def pre_P2R_collect_parameters(self):
        return {}

    def prepare_trainable_params(self):
        self.trainable_params = self.model.get_trainable_params()


class TestRollout:
    def __init__(
        self, name, rollout_world_size, policies_comm, freeze_params: bool = False
    ):
        self.local_rank = int(os.environ.get("LOCAL_RANK", 0))
        self.device = torch.device(f"cuda:{self.local_rank}")
        self.global_rank = int(os.environ.get("RANK", 0))
        self.role = Role.ROLLOUT
        self.world_size = rollout_world_size
        self.policy_to_rollout_nccl_communicators = policies_comm
        rollout_parallelism_config = ParallelismConfig(
            dp_shard_size=1, cp_size=1, tp_size=4, pp_size=1
        )
        self.replica_name = name
        self.parallel_dims = ParallelDims.from_config(
            rollout_parallelism_config,
        )
        self.parallel_dims.build_mesh(device_type="cuda")
        self.model = TestModel(self.device, self.parallel_dims, freeze_params)
        self.parallel_mapper = ParallelTopoMapperGroup(
            self.parallel_dims,
            self.model.config,
            False,
            self.model,
            self.model.weight_mapper,
        )
        self.weight_mapper = self.parallel_mapper.weight_mapper
        compatibale_map = self.model.sharded_tensors
        compatibale_list = self.model.sorted_sharded_params
        operate_compatibale_map = {
            k: torch.zeros(v.shape, dtype=v.dtype).to(self.device)
            for k, v in compatibale_map.items()
        }
        self.ref_compatibale_map = compatibale_map
        self.quantization_type = None
        self.config = CosmosConfig()

        self.vllm_weight_inplace_view_map = compatibale_map
        self.recv_param_key_n_rank_list = compatibale_list
        self.vllm_quantized_weight_map = {}
        self.vllm_hp_weight_map = {}

        self.operate_compatibale_map = operate_compatibale_map
        self.inference_stream = torch.cuda.Stream()
        self.state = State()

        self.recv_weight_shard = types.MethodType(
            vLLMRolloutWorker.recv_weight_shard, self
        )
        # just for testing
        tokenizer = AutoTokenizer.from_pretrained(self.config.policy.model_name_or_path)
        # change the default parallelism config
        self.config.rollout.parallelism.tp_size = 4
        self.config.rollout.parallelism.pp_size = 1

        self.consume_command = types.MethodType(vLLMRolloutWorker.consume_command, self)

        self.rollout = vLLMRollout(self.config, tokenizer)

        self.prepare_trainable_params()

    def get_underlying_model(self):
        return None

    def policy_to_rollout_unicast(self, command: PolicyToRolloutUnicastCommand):
        pass

    def prepare_trainable_params(self):
        self.trainable_params = self.model.get_trainable_params()


async def generate_send_recv_insts(model: TestModel, is_send: bool, global_rank: int):
    policy_parallelism_config = ParallelismConfig(
        dp_shard_size=2, cp_size=1, tp_size=2, pp_size=1
    )
    rollout_parallelism_config = ParallelismConfig(
        dp_shard_size=1, cp_size=1, tp_size=4, pp_size=1
    )
    p_world_size = 4
    r_world_size = 4

    policy_parallel_dims = ParallelDims.from_config_for_analysis(
        policy_parallelism_config, p_world_size
    )
    rollout_parallel_dims = ParallelDims.from_config_for_analysis(
        rollout_parallelism_config, r_world_size
    )

    policy_weight_mapper = GPTWeightMapper(hf_config=model.config)
    rollout_weight_mapper = GPTWeightMapper(hf_config=model.config)

    def dummy(*args, **kwargs):
        return None

    ParallelTopoMapper.parallelism_info_for_dtensor_params = dummy
    ParallelTopoMapper.parallelism_info_for_vllm_params = dummy

    policy_mapper = ParallelTopoMapperGroup(
        global_parallelism=policy_parallel_dims,
        hf_config=model.config,
        is_policy=True,
        underlying_model=None,
        weight_mapper=policy_weight_mapper,
    )
    rollout_mapper = ParallelTopoMapperGroup(
        global_parallelism=rollout_parallel_dims,
        hf_config=model.config,
        is_policy=False,
        underlying_model=None,
        weight_mapper=rollout_weight_mapper,
    )

    def name_to_hf(name: str) -> str:
        return name

    policy_mapper.mapper_group[0].parallelism_info_for_params = {}
    for k, v in model.parallel_spec:
        policy_mapper.mapper_group[0].insert_to_parallelism_info(
            param_name=k, dims_map=v | {"dp_shard_cp": 0}, name_to_hf=name_to_hf
        )

    rollout_mapper.mapper_group[0].parallelism_info_for_params = {}
    for k, v in model.parallel_spec:
        rollout_mapper.mapper_group[0].insert_to_parallelism_info(
            param_name=k,
            dims_map=v | {"dp_shard_cp": 0},
            name_to_hf=name_to_hf,
        )

    local_shards_p = [
        policy_mapper.prepare_local_shard_infos(
            hf_key_n_rank=model.sorted_hf_key_n_rank, global_rank=p_rank
        )
        for p_rank in range(p_world_size)
    ]
    local_shards_r = [
        rollout_mapper.prepare_local_shard_infos(
            hf_key_n_rank=model.sorted_hf_key_n_rank, global_rank=r_rank
        )
        for r_rank in range(r_world_size)
    ]
    cur_dir = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(cur_dir, "configs", "test_simple_grpo.toml")
    with open(config_path, "r") as f:
        config_dict = toml.load(f)
    cosmos_config = CosmosConfig.from_dict(config_dict)
    cosmos_config.policy.parallelism = policy_parallelism_config
    cosmos_config.rollout.parallelism = rollout_parallelism_config
    generator = ParallelizedShardMapper.get_instance(cosmos_config)
    p_params = [[x[0] for x in model.sorted_hf_key_n_rank] for _ in range(p_world_size)]
    r_params = [[x[0] for x in model.sorted_hf_key_n_rank] for _ in range(r_world_size)]
    p_body = {
        "shard_infos": local_shards_p,
        "param_groups": [],
        "sorted_params": p_params,
        "trainable_params": list(model.get_trainable_params()),
    }
    p_data = msgpack.packb(p_body)
    r_body = {
        "shard_infos": local_shards_r,
        "param_groups": [],
        "sorted_params": r_params,
    }
    r_data = msgpack.packb(r_body)

    await generator.set_shard_infos_of_policy(p_data, p_world_size)
    await generator.set_shard_infos_of_rollout(r_data, r_world_size)
    await generator.scheme_generation_done.wait()
    if is_send:
        insts_meta = await generator.get_send_insts_for_policy(global_rank)
    else:
        insts_meta = await generator.get_recv_insts_for_rollout(global_rank)
    insts = msgpack.unpackb(insts_meta, strict_map_key=False)
    policy_to_rollout_insts = [
        WeightSyncInstructionsGroup.from_dict(inst) for inst in insts
    ]
    return policy_to_rollout_insts


async def run_policy_send_to_rollout(shm_name, shm_size, rank, trainable_param_sync):
    """Run as a test policy process to send to rollout process"""
    # Set up NCCL communicator
    policy_name = "policy"
    rollout_name = "rollout"

    # Attach to shared memory
    shm = shared_memory.SharedMemory(name=shm_name)

    command = PolicyToRolloutUnicastCommand(
        policy_name,
        rollout_name,
        POLICY_WORLD_SIZE,
        ROLLOUT_WORLD_SIZE,
        trainable_only=trainable_param_sync,
    )

    try:
        if rank == 0:
            nccl_uid = create_nccl_uid()
            # Create shared memory for NCCL UID
            uid_array = np.ndarray((shm_size + 1,), dtype=np.int64, buffer=shm.buf)
            # Copy NCCL UID to shared memory
            uid_array[:-1] = nccl_uid
            uid_array[-1] = 1
        else:
            uid_array = np.ndarray((shm_size + 1,), dtype=np.int64, buffer=shm.buf)
            while uid_array[-1] == 0:
                time.sleep(0.001)
            assert uid_array[-1] == 1, "Sender process did not set UID correctly"
            nccl_uid = uid_array[:-1].tolist()

        # Create NCCL communicator after UID is shared
        comm_idx = create_nccl_comm(
            nccl_uid, rank, POLICY_WORLD_SIZE + ROLLOUT_WORLD_SIZE
        )
        policy = TestPolicy(
            policy_name,
            POLICY_WORLD_SIZE,
            {policy_name + "_" + rollout_name: comm_idx},
            trainable_param_sync,
        )
        policy.policy_to_rollout_insts = await generate_send_recv_insts(
            policy.model, True, rank
        )
        policy.execute_policy_to_rollout_unicast = types.MethodType(
            GRPOTrainer.execute_policy_to_rollout_unicast, policy
        )
        policy.execute_policy_to_rollout_unicast(command)
        policy.train_stream.synchronize()

    finally:
        # Detach from shared memory
        shm.close()


async def run_rollout_recv_from_policy(shm_name, shm_size, rank, trainable_param_sync):
    """Run as a rollout process to receive from policy process"""
    # Set up NCCL communicator
    policy_name = "policy"
    rollout_name = "rollout"

    # Attach to shared memory
    shm = shared_memory.SharedMemory(name=shm_name)

    command = PolicyToRolloutUnicastCommand(
        policy_name,
        rollout_name,
        POLICY_WORLD_SIZE,
        ROLLOUT_WORLD_SIZE,
        trainable_only=trainable_param_sync,
    )
    try:
        # Get NCCL UID from shared memory
        uid_array = np.ndarray((shm_size + 1,), dtype=np.int64, buffer=shm.buf)
        while uid_array[-1] == 0:
            time.sleep(0.001)
        assert uid_array[-1] == 1, "Sender process did not set UID correctly"
        nccl_uid = uid_array[:-1].tolist()
        # Create NCCL communicator with shared UID
        comm_idx = create_nccl_comm(
            nccl_uid, rank + POLICY_WORLD_SIZE, POLICY_WORLD_SIZE + ROLLOUT_WORLD_SIZE
        )

        rollout = TestRollout(
            rollout_name,
            ROLLOUT_WORLD_SIZE,
            {policy_name + "_" + rollout_name: comm_idx},
            trainable_param_sync,
        )
        rollout.policy_to_rollout_recv_insts = await generate_send_recv_insts(
            rollout.model, False, rank
        )
        rollout.policy_to_rollout_unicast = types.MethodType(
            vLLMRolloutWorker.policy_to_rollout_unicast, rollout
        )
        rollout.prepare_shard_infos_for_weight_sync_insts = lambda: None
        rollout.policy_to_rollout_unicast(command)
        rollout.inference_stream.synchronize()

        for k, v in rollout.operate_compatibale_map.items():
            torch.allclose(v, rollout.ref_compatibale_map[k])

    finally:
        # Detach from shared memory
        shm.close()


def policy_to_policy_sync_common(
    shm_names,
    shm_size,
    rank,
    send,
    nccl_rank,
    nccl_size,
    policy_name,
    replica_name_to_rank,
    command,
):
    """Run as a policy process to perform unicast to another policy process or broadcast to all policy processes"""
    # Attach to shared memory
    shm_names = shm_names.split(",")
    shm_name = shm_names[rank]
    shm = shared_memory.SharedMemory(name=shm_name)

    try:
        # Get NCCL UID from shared memory
        if send:
            nccl_uid = create_nccl_uid()
            # Create shared memory for NCCL UID
            uid_array = np.ndarray((shm_size + 1,), dtype=np.int64, buffer=shm.buf)
            # Copy NCCL UID to shared memory
            uid_array[:-1] = nccl_uid
            uid_array[-1] = 1
        else:
            uid_array = np.ndarray((shm_size + 1,), dtype=np.int64, buffer=shm.buf)
            # Wait for sender process to set UID
            while uid_array[-1] == 0:
                time.sleep(0.001)
            assert uid_array[-1] == 1, "Sender process did not set UID correctly"
            nccl_uid = uid_array[:-1].tolist()

        # Create NCCL communicator with shared UID
        comm_idx = create_nccl_comm(nccl_uid, nccl_rank, nccl_size)

        # Construct the model and trainer
        cur_dir = os.path.dirname(os.path.abspath(__file__))
        config_path = os.path.join(cur_dir, "configs", "test_simple_grpo.toml")

        with open(config_path, "r") as f:
            config_dict = toml.load(f)

        cosmos_config = CosmosConfig.from_dict(
            config_dict,
        )
        parallel_dims = ParallelDims.from_config(cosmos_config.policy.parallelism)
        parallel_dims.build_mesh(device_type="cuda")

        def dummy(self, *args, **kwargs):
            pass

        def dummy_init_nccl(self, replica_name, global_rank, controller_hosts):
            pass

        HighAvailabilitylNccl.__init__ = dummy_init_nccl

        class FakeNCCL:
            def __init__(self, comm_idx):
                self.comm_idx = comm_idx

            def get_replica_rank(self, replica_name: str):
                if replica_name in replica_name_to_rank:
                    return replica_name_to_rank[replica_name]
                else:
                    raise ValueError(
                        f"Replica name {replica_name} not found in mapping."
                    )

            def broadcast(self, tensor: torch.Tensor, src_replica: str):
                src_rank = self.get_replica_rank(src_replica)
                nccl_broadcast(tensor, src_rank, self.comm_idx)

            def send(self, tensor: torch.Tensor, dst_replica: str):
                dst_rank = self.get_replica_rank(dst_replica)
                nccl_send(tensor, dst_rank, self.comm_idx)

            def recv(self, tensor: torch.Tensor, src_replica: str):
                src_rank = self.get_replica_rank(src_replica)
                nccl_recv(tensor, src_rank, self.comm_idx)

            def shutdown(self):
                pass

        Trainer.init_comm = dummy
        CommMixin.init_redis = dummy
        CommMixin.start_heartbeat = dummy
        CommMixin.replica_name = policy_name
        CommMixin.remote_hosts = ["localhost:0"]
        CommMixin.shutdown_signal = threading.Event()
        GRPOTrainer.prepare_shard_infos_for_weight_sync_insts = dummy
        policy = GRPOTrainer(cosmos_config, parallel_dims)
        policy.model_load_from_hf()
        policy.replica_name = policy_name
        policy.inter_policy_nccl = FakeNCCL(comm_idx)
        policy.mesh_ready = True
        policy.replica_name_to_rank = replica_name_to_rank

        def sample_tensor():
            sample_tensors = []
            self_state_dict = policy.model.state_dict()
            sample_tensors.append(self_state_dict[sorted(self_state_dict.keys())[0]])
            sample_tensors.append(self_state_dict[sorted(self_state_dict.keys())[-1]])

            optimizer_state = policy.optimizers.state_dict()
            sample_tensors.append(optimizer_state[sorted(optimizer_state.keys())[0]])
            sample_tensors.append(optimizer_state[sorted(optimizer_state.keys())[-1]])

            lr_sheduler_state = policy.lr_schedulers.state_dict()
            sample_tensors.append(
                lr_sheduler_state[sorted(lr_sheduler_state.keys())[0]]
            )
            sample_tensors.append(
                lr_sheduler_state[sorted(lr_sheduler_state.keys())[-1]]
            )
            sample_tensors = [
                tensor.to_local().cpu()
                if isinstance(tensor, torch.distributed.tensor.DTensor)
                else tensor.cpu()
                if isinstance(tensor, torch.Tensor)
                else tensor
                for tensor in sample_tensors
            ]
            return sample_tensors

        if not send:
            sample_tensors = sample_tensor()

        if isinstance(command, PolicyToPolicyUnicastCommand):
            policy.execute_policy_to_policy_unicast(command)
        elif isinstance(command, PolicyToPolicyBroadcastCommand):
            policy.execute_policy_to_policy_broadcast(command)

        if not send:
            origin_sample_tensors = sample_tensors
            sample_tensors = sample_tensor()
            for tensor, origin_tensor in zip(sample_tensors, origin_sample_tensors):
                if isinstance(tensor, torch.Tensor):
                    assert torch.allclose(
                        tensor, origin_tensor
                    ), f"Tensor values do not match {tensor} {origin_tensor}"
                elif isinstance(tensor, bool):
                    assert (
                        tensor == origin_tensor
                    ), f"Tensor values do not match {tensor} {origin_tensor}"
    finally:
        # Detach from shared memory
        shm.close()


def run_policy_unicast_to_policy(shm_names, shm_size, rank, send):
    """Run as a policy process to perform unicast to another policy process"""
    policy_src_name = "policy_src"
    policy_dst_name = "policy_dst"
    command = PolicyToPolicyUnicastCommand(policy_src_name, policy_dst_name)
    nccl_rank = 0 if send else 1
    nccl_size = 2
    replica_name_to_rank = {policy_src_name: 0, policy_dst_name: 1}
    policy_name = policy_src_name if send else policy_dst_name
    # Call the common function to handle both send and receive
    policy_to_policy_sync_common(
        shm_names,
        shm_size,
        rank,
        send,
        nccl_rank,
        nccl_size,
        policy_name,
        replica_name_to_rank,
        command,
    )


def run_policy_broadcast_to_policy(shm_names, shm_size, rank, total_rep, self_rep):
    """Run as a policy process to perform broadcast to all policy processes"""
    policy_name = "policy_" + str(self_rep)
    policy_src = "policy_0"
    policy_dsts = ["policy_" + str(rep) for rep in range(total_rep)]
    command = PolicyToPolicyBroadcastCommand(policy_src, policy_dsts)
    nccl_rank = self_rep
    nccl_size = total_rep
    replica_name_to_rank = {"policy_" + str(i): i for i in range(total_rep)}
    send = policy_src == policy_name
    # Call the common function to handle both send and receive
    policy_to_policy_sync_common(
        shm_names,
        shm_size,
        rank,
        send,
        nccl_rank,
        nccl_size,
        policy_name,
        replica_name_to_rank,
        command,
    )


def run_overfitting_policy():
    from cosmos_rl.policy.train import main as policy_main
    from cosmos_rl.utils.ulysses import slice_inputs_for_ulysses

    N_STEPS = 30
    training_loss = []

    def train(self):
        def _log_in_master(msg):
            if (
                self.config.logging.logger
                and util.is_master_rank(self.parallel_dims, self.global_rank)
                and "console" in self.config.logging.logger
            ):
                logger.info(msg)

        global_batch = next(iter(self.train_data_loader))
        raw_batch = global_batch[0 : self.config.train.train_policy.mini_batch]

        max_len = min(
            self.config.policy.model_max_length,
            self.data_packer.sft_compute_max_len(raw_batch),
        )

        if self.seq_len_multiple > 1:
            max_len = (
                (max_len + self.seq_len_multiple - 1)
                // self.seq_len_multiple
                * self.seq_len_multiple
            )
        batch = self.data_packer.sft_collate_fn(
            raw_batch,
            computed_max_len=max_len,
            pad_token_id=self.tokenizer.pad_token_id,
            ignore_label_id=-100,
        )

        for k, v in batch.items():
            batch[k] = v.to(self.device) if isinstance(v, torch.Tensor) else v

        labels = batch.pop("label_ids")

        position_ids, input_ids, _ = self.model.get_position_ids(**batch)

        batch["position_ids"] = position_ids
        padding_mask = batch.get("padding_mask", None)

        if self.parallel_dims.cp_enabled:
            [input_ids, position_ids, padding_mask] = slice_inputs_for_ulysses(
                [input_ids, position_ids, padding_mask],
                self.parallel_dims.mesh["cp"],
            )

            batch["input_ids"] = input_ids
            batch["position_ids"] = position_ids
            if padding_mask is not None:
                batch["padding_mask"] = padding_mask

        assert not self.parallel_dims.pp_enabled

        for step in range(N_STEPS):
            _log_in_master(f"Training step {step + 1}/{N_STEPS}")

            self.optimizers.zero_grad()
            self.model.train()
            logits = self.model(**batch)
            loss = self.loss_fn(
                logits,
                labels,
            )
            loss.backward()
            acc_loss = loss.detach()

            """
            Compute the global grad norm on all parameters and then apply
            gradient clipping using the global grad norm.
            """
            grad_norm = None
            if self.config.train.optm_grad_norm_clip > 0:
                # Must pass empty list even if model_part is None,
                # GradNorm across pp stages will fail if some rank does not join the barrier
                all_params = [
                    p
                    for m in [model for model in self.model_parts if model is not None]
                    for p in m.parameters()
                ]
                grad_norm = dist_util.gradient_norm_clipping(
                    all_params,
                    self.config.train.optm_grad_norm_clip,
                    foreach=True,
                    pp_mesh=self.parallel_dims.mesh["pp"]
                    if self.parallel_dims.pp_enabled
                    else None,
                )

            self.optimizers.step()
            self.lr_schedulers.step()

            if (
                self.parallel_dims.dp_replicate_enabled
                or self.parallel_dims.dp_shard_enabled
                or self.parallel_dims.cp_enabled
            ):
                global_avg_loss, _ = (
                    dist_util.dist_mean(acc_loss, self.parallel_dims.mesh["dp_cp"]),
                    dist_util.dist_max(acc_loss, self.parallel_dims.mesh["dp_cp"]),
                )
            else:
                global_avg_loss = acc_loss.item()

            _log_in_master(
                f"Step: {step}/{N_STEPS}, Loss: {global_avg_loss:.5f}, Grad norm: {grad_norm:.5f}, Learning rate: {self.lr_schedulers.get_last_lr()[0]:.5e}"
            )
            training_loss.append(global_avg_loss)

        self.unregister_from_controller()

    SFTTrainer.train = train

    policy_main()

    # check the loss has been decreasing over time
    x = np.arange(len(training_loss))
    m, _ = np.polyfit(x, training_loss, 1)
    print(f"slope is {m}")
    assert m < 0


def run_dummy_policy():
    """Run as a dummy policy process for testing"""
    from cosmos_rl.policy.train import main as policy_main

    def dummy_train_grpo(
        self, current_step: int, total_steps: int, remain_samples_num: int
    ):
        return {}

    def dummy(self):
        pass

    def dummy_model_load_from_hf(self):
        self.model_ready = True

    def dummy_execute_policy_to_rollout_unicast(self, command):
        return False

    GRPOTrainer.train = dummy_train_grpo
    GRPOTrainer.model_load_from_hf = dummy_model_load_from_hf

    def get_policy_command_handler(cls, command_type):
        if command_type == PolicyToRolloutUnicastCommand:
            return dummy_execute_policy_to_rollout_unicast
        return cls.policy_command_handler_registry.get_command_handler(command_type)

    GRPOTrainer.prepare_shard_infos_for_weight_sync_insts = dummy
    GRPOTrainer.get_policy_command_handler = get_policy_command_handler
    SFTTrainer.train = dummy
    policy_main()


def run_dummy_rollout():
    """Run as a dummy rollout process for testing purposes"""
    from cosmos_rl.rollout.rollout_entrance import run_rollout

    def dummy_sync_weight_from_policy(self, command):
        self.state.set_weight_synced()

    def dummy_rollout2rollout_broadcast(self, broadcast_command):
        if broadcast_command.replica_should_stop():
            self.shutdown_signal.set()

    def dummy(self):
        pass

    def get_rollout_command_handler(cls, command_type):
        if command_type == PolicyToRolloutUnicastCommand:
            return dummy_sync_weight_from_policy
        elif command_type == PolicyToPolicyUnicastCommand:
            return dummy_rollout2rollout_broadcast
        return cls.rollout_command_handler_registry.get_command_handler(command_type)

    vLLMRolloutWorker.get_rollout_command_handler = get_rollout_command_handler
    vLLMRolloutWorker.prepare_shard_infos_for_weight_sync_insts = dummy

    def dummy_init(self, config, tokenizer, **kwargs):
        class Llm_engine:
            def step(self, *args, **kwargs):
                pass

        class Rollout_engine:
            llm_engine = Llm_engine()

        self.rollout_engine = Rollout_engine()
        self.eos_token_ids = [0]
        self._engine_initialized = True

        def rollout_generation(
            self,
            prompt_id_and_payload_list,
            stream,
            data_packer,
            sampling_params,
        ):
            payloads = [x[1] for x in prompt_id_and_payload_list]
            completions_per_prompt = [[x] for x in payloads]
            return completions_per_prompt

        self.rollout_generation = types.MethodType(rollout_generation, self)

    vLLMRollout.__init__ = dummy_init
    run_rollout()


def run_policy_parallelism_extract(rank, fsdp, tp, pp):
    cur_dir = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(
        cur_dir,
        "configs",
        "test_simple_grpo.toml",
    )
    with open(config_path, "r") as f:
        config_dict = toml.load(f)
    config = CosmosConfig.from_dict(
        config_dict,
    )
    config.policy.parallelism.dp_shard_size = fsdp
    config.policy.parallelism.tp_size = tp
    config.policy.parallelism.pp_size = pp
    hf_config = util.retry(AutoConfig.from_pretrained)(
        config.policy.model_name_or_path,
        trust_remote_code=True,
    )
    parallel_dims = ParallelDims.from_config(config.policy.parallelism)
    parallel_dims.build_mesh(device_type="cuda")
    model = ModelRegistry.build_model(config)
    try:
        # Apply parallelism to the model
        parallelize_fn, _ = model.parallelize_fn
        parallelize_fn(model, parallel_dims, config, pp_loss_fn=GRPOTrainer.pp_loss_fn)
    except Exception as e:
        import traceback

        traceback.print_exc()
        raise e

    mapper = ParallelTopoMapperGroup(
        parallel_dims,
        hf_config=hf_config,
        is_policy=True,
        underlying_model=model,
        weight_mapper=model.weight_mapper,
    )

    assert len(mapper.mapper_group) == 1, "Only one mapper group expected"
    keys_n_ranks = []
    for name, tensor_or_callable in model.weight_sync_transforms:
        if isinstance(tensor_or_callable, torch.Tensor):
            keys_n_ranks.append((name, tensor_or_callable.ndim))
        else:
            tensor_or_callable = tensor_or_callable()
            keys_n_ranks.append((name, tensor_or_callable.ndim))
    hf_key_n_rank = sorted(keys_n_ranks, key=lambda x: x[0])
    local_shard_infos = mapper.prepare_local_shard_infos(hf_key_n_rank, rank)
    all_rank_local_shard_infos = dist_util.all_gather_object_cpu(local_shard_infos)
    if rank == 0:
        name = config_path = os.path.join(
            cur_dir, "data", f"test_policy_extract_pp_{pp}_fsdp_{fsdp}_tp_{tp}.npy"
        )
        gt = np.load(name, allow_pickle=True)
        np.testing.assert_array_equal(
            np.array(all_rank_local_shard_infos, dtype=object), gt
        )


def run_rollout_parallelism_extract(rank, fsdp, tp, pp):
    cur_dir = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(
        cur_dir,
        "configs",
        "test_simple_grpo.toml",
    )
    with open(config_path, "r") as f:
        config_dict = toml.load(f)
    config = CosmosConfig.from_dict(
        config_dict,
    )
    config.rollout.parallelism.tp_size = tp
    config.rollout.parallelism.pp_size = pp

    hf_config = util.retry(AutoConfig.from_pretrained)(
        config.policy.model_name_or_path,
        trust_remote_code=True,
    )
    tokenizer = util.retry(AutoTokenizer.from_pretrained)(
        config.policy.model_name_or_path
    )
    rollout = vLLMRollout(
        config,
        tokenizer=tokenizer,
    )

    rollout.init_engine(seed=config.rollout.seed, load_format="dummy")
    parallel_dims = ParallelDims.from_config(config.rollout.parallelism)
    parallel_dims.build_mesh(device_type="cuda")

    weight_mapper = WeightMapper.get_weight_mapper(hf_config.model_type)(hf_config)
    mapper = ParallelTopoMapperGroup(
        parallel_dims,
        hf_config=hf_config,
        is_policy=False,
        underlying_model=rollout.get_underlying_model(),
        weight_mapper=weight_mapper,
    )

    assert len(mapper.mapper_group) == 1, "Only one mapper group expected"

    recv_param_key_n_rank_list = []
    _, grouped_recv_param_key_n_rank_list = weight_mapper.rollout_prepare_recv(
        rollout.get_underlying_model()
    )
    for group in grouped_recv_param_key_n_rank_list:
        recv_param_key_n_rank_list.extend(group)

    local_shard_infos = mapper.prepare_local_shard_infos(
        recv_param_key_n_rank_list, rank
    )
    all_rank_local_shard_infos = dist_util.all_gather_object_cpu(local_shard_infos)
    if rank == 0:
        name = config_path = os.path.join(
            cur_dir, "data", f"test_rollout_extract_pp_{pp}_fsdp_{fsdp}_tp_{tp}.npy"
        )
        gt = np.load(name, allow_pickle=True)
        np.testing.assert_array_equal(
            np.array(all_rank_local_shard_infos, dtype=object), gt
        )


class TestModelType:
    num_hidden_layers = 12
    num_attention_heads = 32
    num_key_value_heads = 32
    hidden_size = 1024
    num_attention_heads = 32
    model_type = "gpt"


async def parallel_map_check():
    # Create a mock ParallelismConfig object
    policy_parallelism_config = ParallelismConfig(
        dp_shard_size=-1, cp_size=1, tp_size=2, pp_size=1
    )
    rollout_parallelism_config = ParallelismConfig(
        dp_shard_size=-1, cp_size=1, tp_size=4, pp_size=1
    )
    p_world_size = 8
    r_world_size = 4

    policy_parallel_dims = ParallelDims.from_config_for_analysis(
        policy_parallelism_config, p_world_size
    )
    rollout_parallel_dims = ParallelDims.from_config_for_analysis(
        rollout_parallelism_config, r_world_size
    )

    policy_weight_mapper = GPTWeightMapper(
        hf_config=TestModelType  # Assuming a mock config for testing
    )
    rollout_weight_mapper = GPTWeightMapper(
        hf_config=TestModelType  # Assuming a mock config for testing
    )

    def dummy(*args, **kwargs):
        return None

    ParallelTopoMapper.parallelism_info_for_dtensor_params = dummy
    ParallelTopoMapper.parallelism_info_for_vllm_params = dummy

    policy_mapper = ParallelTopoMapperGroup(
        global_parallelism=policy_parallel_dims,
        hf_config=TestModelType,
        is_policy=True,
        underlying_model=None,
        weight_mapper=policy_weight_mapper,
    )
    rollout_mapper = ParallelTopoMapperGroup(
        global_parallelism=rollout_parallel_dims,
        hf_config=TestModelType,
        is_policy=False,
        underlying_model=None,
        weight_mapper=rollout_weight_mapper,
    )

    assert len(policy_mapper.mapper_group) == 1
    assert len(rollout_mapper.mapper_group) == 1

    def name_to_hf(name: str) -> str:
        return name

    layers = [
        ("model.layers.9.input_layernorm.weight", {}),
        ("model.layers.9.mlp.down_proj.weight", {"tp": 1}),
        ("model.layers.9.mlp.gate_proj.weight", {"tp": 0}),
        ("model.layers.9.mlp.up_proj.weight", {"tp": 0}),
        ("model.layers.9.post_attention_layernorm.weight", {}),
        ("model.layers.9.self_attn.k_proj.bias", {"tp": 0}),
        ("model.layers.9.self_attn.k_proj.weight", {"tp": 0}),
        ("model.layers.9.self_attn.o_proj.weight", {"tp": 0}),
        ("model.layers.9.self_attn.q_proj.bias", {"tp": 0}),
        ("model.layers.9.self_attn.q_proj.weight", {"tp": 0}),
        ("model.layers.9.self_attn.v_proj.bias", {"tp": 0}),
        ("model.layers.9.self_attn.v_proj.weight", {"tp": 0}),
        ("lm_head.weight", {"tp": 0}),
        ("model.norm.weight", {}),
        ("model.embed_tokens.weight", {"tp": 0}),
    ]
    policy_mapper.mapper_group[0].parallelism_info_for_params = {}
    layers.sort(key=lambda x: x[0])
    for k, v in layers:
        policy_mapper.mapper_group[0].insert_to_parallelism_info(
            param_name=k,
            dims_map=v | {"dp_shard_cp": 0},
            name_to_hf=name_to_hf,
        )

    rollout_mapper.mapper_group[0].parallelism_info_for_params = {}
    for k, v in layers:
        rollout_mapper.mapper_group[0].insert_to_parallelism_info(
            param_name=k,
            dims_map=v | {"dp_shard_cp": 0},
            name_to_hf=name_to_hf,
        )

    local_shards_p = [
        policy_mapper.prepare_local_shard_infos(
            hf_key_n_rank=layers, global_rank=p_rank
        )
        for p_rank in range(p_world_size)
    ]
    local_shards_r = [
        rollout_mapper.prepare_local_shard_infos(
            hf_key_n_rank=layers, global_rank=r_rank
        )
        for r_rank in range(r_world_size)
    ]

    p_params = [[x[0] for x in layers] for _ in range(p_world_size)]
    r_params = [[x[0] for x in layers] for _ in range(r_world_size)]
    p_body = {
        "shard_infos": local_shards_p,
        "param_groups": [],
        "sorted_params": p_params,
        "trainable_params": [x[0] for x in layers],
    }
    p_data = msgpack.packb(p_body)
    r_body = {
        "shard_infos": local_shards_r,
        "param_groups": [],
        "sorted_params": r_params,
    }
    r_data = msgpack.packb(r_body)

    cur_dir = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(cur_dir, "configs", "test_simple_grpo.toml")
    with open(config_path, "r") as f:
        config_dict = toml.load(f)
    cosmos_config = CosmosConfig.from_dict(config_dict)
    cosmos_config.policy.parallelism = policy_parallelism_config
    cosmos_config.rollout.parallelism = rollout_parallelism_config
    generator = ParallelizedShardMapper.get_instance(cosmos_config)
    await generator.set_shard_infos_of_policy(p_data, p_world_size)
    await generator.set_shard_infos_of_rollout(r_data, r_world_size)

    await generator.scheme_generation_done.wait()
    global_rank = 5
    # generator.rollout_from_policy_insts_meta = [
    #     [{} for _ in range(generator.rollout_world_size)]
    #     for _ in range(generator.policy_world_size)
    # ]

    insts = await generator.get_send_insts_for_policy(global_rank)
    r_rank_max = 0
    layer_idx = 0

    layers.sort(key=lambda x: x[0])
    insts = msgpack.unpackb(insts, strict_map_key=False)
    for inst_group in insts:
        for inst in inst_group["param_instructions"]:
            dest_name = inst["param_name"]
            for i in inst["instructions"]:
                p_rank = i["policy_rank"]
                r_rank = i["rollout_rank"]
                assert p_rank == global_rank
                while layers[layer_idx][0] != dest_name:
                    r_rank_max = 0
                    layer_idx += 1
                assert layers[layer_idx][0] == dest_name
                assert r_rank >= r_rank_max
                if r_rank > r_rank_max:
                    r_rank_max = r_rank

    global_rank = 2
    insts = await generator.get_recv_insts_for_rollout(global_rank)
    insts = msgpack.unpackb(insts, strict_map_key=False)

    p_rank_max = 0
    layer_idx = 0
    for inst_group in insts:
        for inst in inst_group["param_instructions"]:
            dest_name = inst["param_name"]
            for i in inst["instructions"]:
                p_rank = i["policy_rank"]
                r_rank = i["rollout_rank"]
                assert r_rank == global_rank
                while layers[layer_idx][0] != dest_name:
                    p_rank_max = 0
                    layer_idx += 1
                assert layers[layer_idx][0] == dest_name
                assert p_rank >= p_rank_max
                if p_rank > p_rank_max:
                    p_rank_max = p_rank


def run_sft_for_sequence_packing(fsdp, tp, cp):
    def train_test(self, packing_seq):
        train_dataset, _ = construct_dataset(
            config.train.train_policy,
            tokenizer=self.tokenizer,
            data_packer=self.data_packer,
            user_provided_dataset=self.sft_user_dataset,
        )
        train_sampler = DistributedSampler(
            train_dataset,
            num_replicas=self.dp_world_size,
            rank=self.dp_rank,
            shuffle=False,
            drop_last=False,
        )

        def get_train_data_loader(sampler: Sampler[int]):
            return DataLoader(
                train_dataset,
                batch_size=config.train.train_batch_per_replica,
                shuffle=False,
                num_workers=config.train.train_policy.dataloader_num_workers,
                prefetch_factor=config.train.train_policy.dataloader_prefetch_factor,
                sampler=sampler,
                collate_fn=collate_fn,
                drop_last=False,
            )

        self.train_data_loader = get_train_data_loader(train_sampler)
        losses = []
        for global_batch in self.train_data_loader:
            acc_loss = torch.zeros(1, device=self.device)
            self.optimizers.zero_grad()
            global_batch_size = len(global_batch)
            # split global_batch into mini_batches
            mini_batch_begin_idxs = list(
                range(
                    0,
                    global_batch_size,
                    self.config.train.train_policy.mini_batch,
                )
            )
            for i in mini_batch_begin_idxs:
                raw_batch = global_batch[
                    i : i + self.config.train.train_policy.mini_batch
                ]
                max_len = min(
                    self.config.policy.model_max_length,
                    self.data_packer.sft_compute_max_len(raw_batch),
                )
                if self.seq_len_multiple > 1:
                    max_len = (
                        (max_len + self.seq_len_multiple - 1)
                        // self.seq_len_multiple
                        * self.seq_len_multiple
                    )
                batch = self.data_packer.sft_collate_fn(
                    raw_batch,
                    computed_max_len=max_len,
                    pad_token_id=self.tokenizer.pad_token_id,
                    ignore_label_id=-100,
                )
                self.model.train()
                for k, v in batch.items():
                    batch[k] = v.to(self.device) if isinstance(v, torch.Tensor) else v
                labels = batch.pop("label_ids")
                position_ids, input_ids, pos_seq_dim = self.model.get_position_ids(
                    **batch
                )
                batch["position_ids"] = position_ids
                padding_mask = batch.get("padding_mask", None)
                if packing_seq:
                    # Prepare for the sequence packing information.
                    packed_args = pack_sequences_info_collect(
                        batch["input_ids"],
                        pad_token_id=self.tokenizer.pad_token_id,
                        label_ids=labels,
                        ignore_label_id=-100,
                        seq_len_multiple=self.seq_len_multiple,
                    )
                    batch.update(packed_args)
                    labels = pack_sequences_for_labels(labels, batch["valid_input_len"])
                    packed_args = pack_sequences_for_masks(
                        batch["valid_input_len"], batch["valid_input_len"]
                    )
                    batch.update(packed_args)

                if self.parallel_dims.cp_enabled and not packing_seq:
                    [input_ids, position_ids, padding_mask] = slice_inputs_for_ulysses(
                        [input_ids, position_ids, padding_mask],
                        self.parallel_dims.mesh["cp"],
                        seq_dims=[1, pos_seq_dim, 1],
                    )
                    batch["input_ids"] = input_ids
                    batch["position_ids"] = position_ids
                    if padding_mask is not None:
                        batch["padding_mask"] = padding_mask

                if self.parallel_dims.cp_enabled and packing_seq:
                    # Slice for cp after embedding generation and sequence packing in the model forward later.
                    batch["cp_mesh"] = self.parallel_dims.mesh["cp"]
                logits = self.model(**batch)

                loss = self.loss_fn(
                    logits,
                    labels,
                    output_packing_mask=batch.get("input_packing_mask", None),
                    target_packing_mask=batch.get("label_packing_mask", None),
                    loss_scaling_factor=1.0 / len(mini_batch_begin_idxs),
                )
                loss.backward()
                acc_loss += loss.detach()
                all_params = [
                    p
                    for m in [model for model in self.model_parts if model is not None]
                    for p in m.parameters()
                ]
                dist_util.gradient_norm_clipping(
                    all_params,
                    self.config.train.optm_grad_norm_clip,
                    foreach=True,
                    pp_mesh=self.parallel_dims.mesh["pp"]
                    if self.parallel_dims.pp_enabled
                    else None,
                    return_norm_only=(self.config.train.optm_grad_norm_clip <= 0.0),
                )
                self.optimizers.step()
                self.lr_schedulers.step()
                self.train_step += 1
                if (
                    self.parallel_dims.dp_replicate_enabled
                    or self.parallel_dims.dp_shard_enabled
                    or self.parallel_dims.cp_enabled
                ):
                    global_avg_loss, global_max_loss = (  # noqa: F841
                        dist_util.dist_mean(acc_loss, self.parallel_dims.mesh["dp_cp"]),
                        dist_util.dist_max(acc_loss, self.parallel_dims.mesh["dp_cp"]),
                    )
                else:
                    global_avg_loss = global_max_loss = acc_loss.item()  # noqa: F841

                if util.is_master_rank(self.parallel_dims, self.global_rank):
                    losses.append(global_avg_loss)
                if self.train_step >= 8:
                    return losses
        return losses

    cur_dir = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(
        cur_dir,
        "configs",
        "test_simple_sft.toml",
    )
    with open(config_path, "r") as f:
        config_dict = toml.load(f)
    config = CosmosConfig.from_dict(
        config_dict,
    )
    config.policy.parallelism.dp_shard_size = fsdp
    config.policy.parallelism.tp_size = tp
    config.policy.parallelism.cp_size = cp
    logger.info(f"[Test] sequence packing with fsdp {fsdp}, tp {tp}, cp {cp}")
    parallel_dims = ParallelDims.from_config(
        parallesim_config=config.policy.parallelism
    )
    init_distributed()
    parallel_dims.build_mesh(device_type="cuda")

    def dummy(self):
        self.replica_name = str(dist_utils.broadcast_object_cpu(uuid.uuid4()))
        # `sft_user_dataset` is only used in SFT mode when the user provides a dataset
        self.sft_user_dataset = None
        hf_config = util.retry(AutoConfig.from_pretrained)(
            self.config.policy.model_name_or_path, trust_remote_code=True
        )
        model_type = hf_config.model_type
        logger.info(f"model type {model_type}")
        self.data_packer = DecoderOnlyLLMDataPacker()
        self.data_packer.setup(self.config, self.tokenizer)
        self.remote_hosts = ["0.0.0.0:8000"]
        pass

    CommMixin.init_comm = dummy
    trainer = SFTTrainer(config=config, parallel_dims=parallel_dims)
    non_packing_losses = train_test(trainer, False)
    trainer = SFTTrainer(config=config, parallel_dims=parallel_dims)
    packing_losses = train_test(trainer, True)
    if util.is_master_rank(trainer.parallel_dims, trainer.global_rank):
        assert len(non_packing_losses) == 8
        assert len(packing_losses) == 8
        logger.info(f"[Test] non_packing_losses: {non_packing_losses}")
        logger.info(f"[Test] packing_losses: {packing_losses}")
        assert np.allclose(non_packing_losses, packing_losses, atol=1e-3, rtol=1e-3)


async def main():
    # Get shared memory name and size from command line arguments
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--shm_name", type=str)  # 1st arg
    parser.add_argument("--shm_size", type=int)  # 2nd arg
    parser.add_argument("--mode", type=str, required=True)  # 3rd arg
    parser.add_argument(
        "--parallel_config",
        type=str,
        required=False,
        default="",
        help="Parallel dimensions to use for the test. Format: fsdp;tp;pp",
    )
    parser.add_argument(
        "--trainable_param_sync",
        type=bool,
        required=False,
        default=False,
        help="If only trainable params are synced. If set, part of the params will be frozen.",
    )  # 4th arg
    args = parser.parse_args()
    mode = args.mode
    shm_name = args.shm_name
    shm_size = args.shm_size
    mode = args.mode
    trainable_param_sync = (
        args.trainable_param_sync
    )  # If only trainable params are synced

    if mode == "dummy_policy":
        os.environ["COSMOS_ROLE"] = "Policy"
        # Dummy policy process for testing
        run_dummy_policy()
        exit(0)
    elif mode == "dummy_rollout":
        os.environ["COSMOS_ROLE"] = "Rollout"
        # Dummy rollout process for testing
        run_dummy_rollout()
        exit(0)
    elif mode == "test_overfit":
        run_overfitting_policy()
        exit(0)
    elif mode == "sft_for_sequence_packing":
        sepc = args.parallel_config
        fsdp, tp, cp = sepc.split(";")
        fsdp = int(fsdp.split(":")[1])
        tp = int(tp.split(":")[1])
        cp = int(cp.split(":")[1])
        run_sft_for_sequence_packing(fsdp, tp, cp)
        exit(0)

    # Initialize distributed environment
    init_distributed()
    local_rank = int(os.environ.get("LOCAL_RANK", 0))
    device = torch.device(f"cuda:{local_rank}")
    torch.cuda.set_device(device)
    if not dist.is_initialized():
        rank = 0
        world_size = 1
    else:
        rank = dist.get_rank()
        world_size = dist.get_world_size()
    print(f"Rank {rank} started with mode {mode} {torch.cuda.current_device()}")

    if mode == "policy_send_to_rollout":
        assert (
            world_size == POLICY_WORLD_SIZE
        ), "World size must match POLICY_WORLD_SIZE for policy process"
        await run_policy_send_to_rollout(shm_name, shm_size, rank, trainable_param_sync)
    elif mode == "rollout_recv_from_policy":
        assert (
            world_size == ROLLOUT_WORLD_SIZE
        ), "World size must match ROLLOUT_WORLD_SIZE for rollout process"
        await run_rollout_recv_from_policy(
            shm_name, shm_size, rank, trainable_param_sync
        )
    elif mode == "policy_send_to_policy":
        run_policy_unicast_to_policy(shm_name, shm_size, rank, True)
    elif mode == "policy_recv_from_policy":
        run_policy_unicast_to_policy(shm_name, shm_size, rank, False)
    elif mode.startswith("policy_broadcast_to_policy"):
        total_rep = int(mode.split(",")[1])
        self_rep = int(mode.split(",")[2])
        run_policy_broadcast_to_policy(shm_name, shm_size, rank, total_rep, self_rep)
    elif mode == "policy_parallelism_extract":
        sepc = args.parallel_config
        fsdp, tp, pp = sepc.split(";")
        fsdp = int(fsdp.split(":")[1])
        tp = int(tp.split(":")[1])
        pp = int(pp.split(":")[1])
        run_policy_parallelism_extract(rank, fsdp, tp, pp)
    elif mode == "rollout_parallelism_extract":
        sepc = args.parallel_config
        fsdp, tp, pp = sepc.split(";")
        fsdp = int(fsdp.split(":")[1])
        tp = int(tp.split(":")[1])
        pp = int(pp.split(":")[1])
        run_rollout_parallelism_extract(rank, fsdp, tp, pp)
    elif mode == "parallel_map_check":
        await parallel_map_check()
    else:
        raise ValueError("Invalid mode.")
    # Clean up distributed environment
    destroy_distributed()


if __name__ == "__main__":
    asyncio.run(main())



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict, Any, Optional
from dataclasses import dataclass, field
import math
from cosmos_rl.dispatcher.protocol import Role, MESH_NAMES, RegisterRequest
from cosmos_rl.policy.config import SubProfilerConfig
import asyncio
from cosmos_rl.dispatcher.algo.base import RuleBasedAlgo
import weakref
from cosmos_rl.utils.logging import logger
from cosmos_rl.utils.redis_stream import RedisStreamHandler
import msgpack
import time


@dataclass
class Rollout:
    payload: Any
    completion: str
    is_end: bool
    reward: float
    advantage: float
    prompt_idx: int
    n_ignore_prefix_tokens: int = 0

    def __init__(
        self,
        payload: Any,
        completion: str,
        is_end: bool,
        reward: float,
        advantage: float,
        prompt_idx: int,
        n_ignore_prefix_tokens: int = 0,
    ):
        self.payload = payload
        self.completion = completion
        self.is_end = is_end
        self.reward = reward
        self.advantage = advantage
        self.prompt_idx = prompt_idx
        self.n_ignore_prefix_tokens = n_ignore_prefix_tokens

    @classmethod
    def from_dict(cls, dict_v: Dict[str, Any]) -> "Rollout":
        return cls(**dict_v)


class RolloutGroup:
    """
    RolloutGroup is a data structure that contains the prompt and completions of a rollout.
    For MutliModal-LM, image/video/audio could be included in the extra_info.
    """

    def __init__(
        self,
        prompt_idx: int,
        payload: Any,
        completions: List[str],
        is_end: bool,
        reference_answer: str,
    ):
        self.prompt_idx: int = prompt_idx
        self.payload: Any = payload
        self.completions: List[str] = completions
        self.is_end: bool = is_end
        self.reference_answer: str = reference_answer

    def compute_rollouts(self, algo: RuleBasedAlgo) -> List[Rollout]:
        assert (
            self.reference_answer is not None
        ), "[RolloutGroup] Reference answer is not provided"

        rewards = [
            algo.compute_reward(completion, self.reference_answer)
            for completion in self.completions
        ]
        logger.debug(f"[RolloutGroup] Rewards: {rewards}")
        advantages = algo.compute_advantage(rewards)
        logger.debug(f"[RolloutGroup] Advantages: {advantages}")

        return [
            Rollout(
                payload=self.payload,
                completion=completion,
                is_end=self.is_end,
                reward=reward,
                advantage=advantage,
                prompt_idx=self.prompt_idx,
            )
            for completion, reward, advantage in zip(
                self.completions, rewards, advantages
            )
        ]


class BatchedRolloutGroup:
    """
    Batched Wrapper of the RolloutGroup
    """

    def __init__(self):
        self.rollout_groups: List[RolloutGroup] = []

    def __len__(self):
        return len(self.rollout_groups)

    def __getitem__(self, idx: int) -> RolloutGroup:
        return self.rollout_groups[idx]

    def __setitem__(self, idx: int, rollout_group: RolloutGroup):
        self.rollout_groups[idx] = rollout_group

    def __delitem__(self, idx: int):
        del self.rollout_groups[idx]

    @classmethod
    def from_rollout_groups(
        cls, rollout_groups: List[RolloutGroup]
    ) -> "BatchedRolloutGroup":
        batched_rollout_group = cls()
        batched_rollout_group.rollout_groups = rollout_groups
        return batched_rollout_group


@dataclass
class Atom:
    """
    Atom is the smallest unit of a computation mesh.
    Usually it is a single GPU process.

    replica_name: str Name of the replica that this atom belongs to
    ranks: List[int] Rank of each dimension in the order of MESH_NAMES
    group_size: List[int] Size of each dimension in the order of MESH_NAMES
    global_rank: int global rank of the atom in this replica.
    host_ip: str IP address of the host machine that this atom belongs to.
    host_name: str Name of the host machine that this atom belongs to.
    trace_path: str Path to the trace file of this atom/rank that stores. May be s3 or local path.
    """

    replica_name: str
    ranks: List[int]
    global_rank: int
    host_ip: str
    host_name: str
    trace_path: Optional[str]
    group_size: List[int]
    rollout_: asyncio.Queue[Rollout]

    def __init__(
        self,
        global_rank: int,
        host_ip: str,
        host_name: str,
        trace_path: str,
        ranks: List[int],
        group_size: List[int],
        replica_name: str,
    ):
        self.ranks = ranks
        self.group_size = group_size
        self.replica_name = replica_name
        self.rollout_queue: asyncio.Queue[Rollout] = asyncio.Queue()
        self._replica: Optional[weakref.ReferenceType["Replica"]] = None
        self.global_rank = global_rank
        self.host_ip = host_ip
        self.host_name = host_name
        self.trace_path = trace_path

    def to_dict(self) -> Dict[str, Any]:
        return {
            "ranks": self.ranks,
            "group_size": self.group_size,
            "replica_name": self.replica_name,
            "global_rank": self.global_rank,
            "host_ip": self.host_ip,
            "host_name": self.host_name,
            "trace_path": self.trace_path,
            "replica": self._replica().name if self._replica else "None",
        }

    @property
    def replica(self) -> "Replica":
        assert self._replica is not None, f"Atom {self} is not bound to a replica"
        return self._replica()

    def bind_replica(self, replica: "Replica"):
        self._replica = weakref.ref(replica)

    async def put_rollout(self, rollout: Rollout):
        assert (
            self.tp_rank() == 0 and self.cp_rank() == 0 and self.pp_rank() == 0
        ), f"Atom {self} is not a dp_shard master atom, cannot put rollout"
        self.rollout_queue.put_nowait(rollout)

    async def fetch_rollouts(self) -> List[Rollout]:
        assert (
            self.tp_rank() == 0 and self.cp_rank() == 0 and self.pp_rank() == 0
        ), f"Atom {self} is not a dp_shard master atom, cannot fetch rollouts"
        rollouts = []
        while not self.rollout_queue.empty():
            rollouts.append(self.rollout_queue.get_nowait())
        return rollouts

    @classmethod
    def from_register_request(cls, request: RegisterRequest) -> "Atom":
        return cls(
            global_rank=request.global_rank,
            host_ip=request.host_ip,
            host_name=request.host_name,
            trace_path=None,
            ranks=request.ranks,
            group_size=request.group_size,
            replica_name=request.replica_name,
        )

    def tp_rank(self) -> int:
        return self.ranks[MESH_NAMES.index("tp")]

    def cp_rank(self) -> int:
        return self.ranks[MESH_NAMES.index("cp")]

    def pp_rank(self) -> int:
        return self.ranks[MESH_NAMES.index("pp")]

    def dp_shard_rank(self) -> int:
        return self.ranks[MESH_NAMES.index("dp_shard")]

    def __post_init__(self):
        assert (
            len(self.ranks) == len(self.group_size)
        ), f"Ranks and group_size must have the same length, got {len(self.ranks)} and {len(self.group_size)}"
        assert (
            len(self.ranks) == len(MESH_NAMES)
        ), f"Ranks must have the same length as MESH_NAMES, got {len(self.ranks)} and {len(MESH_NAMES)}"

    def __str__(self):
        return f"{self.replica_name}_{self.global_rank}"


@dataclass
class ReplicaStatus:
    """
    Status of a replica.
    """

    heartbeat_timestamp: int  # Timestamp of the last heartbeat
    nccl_error_timestamp: Optional[int]  # Timestamp of the last NCCL error
    mesh_rank: int  # rank in the mesh
    ended: bool = False

    def __init__(self):
        self.heartbeat_timestamp = time.time()
        self.nccl_error_timestamp = None
        self.mesh_rank = -1
        self.ended = False


@dataclass
class Replica:
    """
    Replica is a single `DP Relicate` unit, where sub:
        - TP
        - CP
        - PP
        - DP_SHARD
    could be used to form a `DP Relicate` unit.

    name: str Name of the replica
    role: Role Role of the replica, either POLICY or ROLLOUT
    atoms: Dict[str, Atom] Sub-units of the replica
    """

    name: str
    role: Role
    atoms: Dict[str, Atom]
    command_queue: asyncio.Queue
    weights_loaded_in_view_of_command: bool = False
    start_time: int = -1

    status: ReplicaStatus = field(default_factory=ReplicaStatus)

    # For profiling
    sub_profiler_config: SubProfilerConfig = field(default_factory=SubProfilerConfig)

    def __init__(self, name: str, role: Role, atoms: List[Atom]):
        self.name = name  # Note: name must be unique across all the replicas.
        self.role = role
        self.atoms = {str(atom): atom for atom in atoms}
        self.command_queue = asyncio.Queue()

        self.sub_profiler_config = SubProfilerConfig()
        self.status = ReplicaStatus()

    def to_dict(self) -> Dict[str, Any]:
        atoms = []
        for atom_key, atom in self.atoms.items():
            atoms.append(atom.to_dict())
        return {
            "name": self.name,
            "role": self.role,
            "atoms": atoms,
            "arrived": self.all_atoms_arrived,
        }

    @property
    def in_mesh(self) -> bool:
        return self.status.mesh_rank >= 0

    def get_atom(self, ranks: List[int]) -> Atom:
        assert (
            len(ranks) == len(MESH_NAMES)
        ), f"Ranks must have the same length as MESH_NAMES, got {len(ranks)} and {len(MESH_NAMES)}"
        return self.atoms[str(ranks)]

    def arrive(self, atom: Atom):
        assert str(atom) not in self.atoms, f"Atom {atom} already exists"
        # Verify group size is consistent with existing atoms
        if len(self.atoms) > 0:
            existing_atom = next(iter(self.atoms.values()))
            assert (
                atom.group_size == existing_atom.group_size
            ), f"Atom {atom} has inconsistent group size"
        self.atoms[str(atom)] = atom
        if self.role == Role.POLICY and self.all_atoms_arrived:
            # Check out all the dp_shard atoms
            # Only those atoms are responsible rollout fetching
            #    1. pipeline rank is 0
            #    2. dp_shard rank is any
            #    3. cp rank is 0
            #    4. tp rank is 0
            any_atom = next(iter(self.atoms.values()))
            group_sizes = any_atom.group_size
            assert (
                len(group_sizes) == len(MESH_NAMES)
            ), f"Group sizes must have the same length as MESH_NAMES, got {len(group_sizes)} and {len(MESH_NAMES)}"
            tp_size, cp_size, dp_shard_size, pp_size = 1, 1, 1, 1

            for i in range(len(MESH_NAMES)):
                if MESH_NAMES[i] == "tp":
                    tp_size = group_sizes[i]
                elif MESH_NAMES[i] == "cp":
                    cp_size = group_sizes[i]
                elif MESH_NAMES[i] == "pp":
                    pp_size = group_sizes[i]
                elif MESH_NAMES[i] == "dp_shard":
                    dp_shard_size = group_sizes[i]
                else:
                    raise ValueError(f"Unknown mesh name: {MESH_NAMES[i]}")
            assert (
                tp_size * cp_size * dp_shard_size * pp_size == len(self.atoms)
            ), f"Group sizes must be consistent with the number of atoms, got {tp_size} * {cp_size} * {dp_shard_size} * {pp_size} = {tp_size * cp_size * dp_shard_size * pp_size} and {len(self.atoms)}"

    def n_atoms_per_replica(self) -> int:
        """
        Returns the number of atoms per replica.
        This is the product of all group sizes.
        """
        assert len(self.atoms) > 0, f"Replica {self.name} has no atoms"
        atom = next(iter(self.atoms.values()))
        return math.prod(atom.group_size)

    @property
    def all_atoms_arrived(self) -> bool:
        assert len(self.atoms) > 0, f"Replica {self.name} has no atoms"
        atom = next(iter(self.atoms.values()))
        # Dot product of ranks and group_size should be equal
        return len(self.atoms) == math.prod(atom.group_size)

    def put_rollout(self, rollout: Rollout, redis_handler: RedisStreamHandler):
        assert (
            self.role == Role.POLICY
        ), f"Replica {self.name} is not a policy replica, cannot put rollout"
        assert (
            self.all_atoms_arrived
        ), f"Replica {self.name} tries to put rollout but not all atoms have arrived"
        # Check which atom should handle this rollout
        # Publish the rollout to the redis stream to be consumed by policy replicas
        redis_handler.publish_rollout(msgpack.packb(rollout.__dict__), self.name)

    async def find_atom(self, global_rank: int) -> Atom:
        key = f"{self.name}_{global_rank}"
        return self.atoms.get(key)

    async def set_trace_path(self, trace_path: str, global_rank: int):
        atom = await self.find_atom(global_rank)
        logger.info(f"[Controller]: Set trace path of atom {atom} to {trace_path}")
        atom.trace_path = trace_path
        return str(atom)

    def __eq__(self, other):
        return isinstance(other, Replica) and self.name == other.name

    def __hash__(self):
        return hash(self.name)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cosmos_rl.policy.model.base import WeightMapper
from transformers.models.qwen2_5_vl.configuration_qwen2_5_vl import (
    Qwen2_5_VLConfig,
)
import torch
from cosmos_rl.utils import util
from transformers import AutoConfig
from typing import Dict, List, Tuple
import re


class QwenVL25WeightMapper(WeightMapper):
    def __init__(self, hf_config: AutoConfig):
        super().__init__(hf_config)
        assert isinstance(self.config, Qwen2_5_VLConfig)

        self.kv_head_ratio = (
            self.config.num_attention_heads // self.config.num_key_value_heads
        )
        self.head_dim = self.config.hidden_size // self.config.num_attention_heads

    def _rollout_vllm_name_to_hf(self, rollout_weight_name: str) -> str:
        if self.backend == "vllm":
            # Happen to be the same as policy name mapping.
            return self.policy_map_local_key_to_hf_key(rollout_weight_name)
        elif self.backend == "trtllm":
            if rollout_weight_name.startswith("llm.model."):
                hf_name = rollout_weight_name.replace("llm.model.", "model.")
            elif rollout_weight_name.startswith("llm.lm_head."):
                hf_name = rollout_weight_name.replace("llm.lm_head.", "lm_head.")
            elif rollout_weight_name.startswith("mm_encoder."):
                hf_name = rollout_weight_name.replace(".", "visual.")
            else:
                hf_name = rollout_weight_name
            return hf_name
        else:
            raise ValueError(f"Unsupported backend: {self.backend}")

    def __rollout_split_qkv_weight(self, name, weight: torch.Tensor):
        # visual
        if "visual" in name:
            # split qkv weight for visual
            # weight has shape [3 * head_dim, hidden_dim]
            # kv head ratio is 1, so we can split it into q, k, v
            assert (
                weight.shape[0] % 3 == 0
            ), "Weight shape is not compatible for splitting."
            unit_dim = weight.shape[0] // 3  # for both weight and bias
            q_weight = weight[:unit_dim]
            k_weight = weight[unit_dim : unit_dim * 2]
            v_weight = weight[unit_dim * 2 :]
            return q_weight, k_weight, v_weight
        # language
        # split qkv weight
        # weight has shape [q_num_heads * head_dim + k_num_heads * head_dim + v_num_heads * head_dim, hidden_dim]
        shares = self.kv_head_ratio + 2
        dim_0 = weight.shape[0]  # for both weight and bias
        unit_dim = dim_0 // shares

        q_weight = weight[: unit_dim * self.kv_head_ratio]
        k_weight = weight[
            unit_dim * self.kv_head_ratio : unit_dim * (self.kv_head_ratio + 1)
        ]
        v_weight = weight[unit_dim * (self.kv_head_ratio + 1) :]
        return q_weight, k_weight, v_weight

    def _split_gate_proj_weight(self, name, weight: torch.Tensor):
        # gate_proj and up_proj in vllm is already split.
        # weight has shape [2 * x, hidden_dim]
        dim_0 = weight.shape[0]
        gate_proj_weight = weight[: dim_0 // 2]
        up_proj_weight = weight[dim_0 // 2 :]
        return gate_proj_weight, up_proj_weight

    def rollout_prepare_recv(
        self,
        vllm_model,
    ) -> Tuple[
        Dict[str, torch.Tensor],
        List[List[Tuple[str, torch.Size]]],
    ]:
        recv_key_n_rank_list = []
        vllm_weight_inplace_view_map = {}
        for param_name, param in vllm_model.named_parameters():
            group_keys = []
            compatible_key = self._rollout_vllm_name_to_hf(param_name)

            if "qkv_proj" in compatible_key:
                q_weight, k_weight, v_weight = self.__rollout_split_qkv_weight(
                    compatible_key, param
                )
                q_proj_weight_key = compatible_key.replace("qkv_proj", "q_proj")
                k_proj_weight_key = compatible_key.replace("qkv_proj", "k_proj")
                v_proj_weight_key = compatible_key.replace("qkv_proj", "v_proj")
                vllm_weight_inplace_view_map[q_proj_weight_key] = q_weight
                group_keys.append((q_proj_weight_key, q_weight.ndim))
                vllm_weight_inplace_view_map[k_proj_weight_key] = k_weight
                group_keys.append((k_proj_weight_key, k_weight.ndim))
                vllm_weight_inplace_view_map[v_proj_weight_key] = v_weight
                group_keys.append((v_proj_weight_key, v_weight.ndim))
            elif "gate_up_proj" in compatible_key:
                # split gate and up proj
                gate_proj_weight, up_proj_weight = self._split_gate_proj_weight(
                    compatible_key, param
                )
                gate_proj_weight_key = compatible_key.replace(
                    "gate_up_proj", "gate_proj"
                )
                vllm_weight_inplace_view_map[gate_proj_weight_key] = gate_proj_weight
                group_keys.append((gate_proj_weight_key, gate_proj_weight.ndim))

                up_proj_weight_key = compatible_key.replace("gate_up_proj", "up_proj")
                vllm_weight_inplace_view_map[up_proj_weight_key] = up_proj_weight
                group_keys.append((up_proj_weight_key, up_proj_weight.ndim))
            elif "qkv" in compatible_key and "visual" in compatible_key:
                q_weight, k_weight, v_weight = self.__rollout_split_qkv_weight(
                    compatible_key, param
                )
                q_visual_proj_weight_key = compatible_key.replace("qkv", "q")
                k_visual_proj_weight_key = compatible_key.replace("qkv", "k")
                v_visual_proj_weight_key = compatible_key.replace("qkv", "v")
                vllm_weight_inplace_view_map[q_visual_proj_weight_key] = q_weight
                group_keys.append((q_visual_proj_weight_key, q_weight.ndim))
                vllm_weight_inplace_view_map[k_visual_proj_weight_key] = k_weight
                group_keys.append((k_visual_proj_weight_key, k_weight.ndim))
                vllm_weight_inplace_view_map[v_visual_proj_weight_key] = v_weight
                group_keys.append((v_visual_proj_weight_key, v_weight.ndim))
            else:
                vllm_weight_inplace_view_map[compatible_key] = param
                group_keys.append((compatible_key, param.ndim))
            recv_key_n_rank_list.append(group_keys)
        return vllm_weight_inplace_view_map, recv_key_n_rank_list

    def policy_map_local_key_to_hf_key(self, name: str) -> str:
        name = util.clear_weight_name(name)
        if name.startswith("language_model."):
            name = name.replace("language_model.", "")
        if name == "model.lm_head.weight":
            name = "lm_head.weight"
        return name

    def name_to_model_part_index(self, dest_name: str) -> int:
        if dest_name in ["lm_head.weight", "lm_head.bias"]:
            return 0
        elif dest_name.startswith("visual."):
            return 1
        elif dest_name.startswith("model."):
            return 0
        else:
            raise ValueError(f"Unsupported weight: {dest_name}")

    def policy_decompose_param_1_to_n_for_sync(self, name):
        if match := re.search(  # noqa: F841
            r"visual\.blocks\.(\d+)\.attn\.qkv\.(weight|bias)",
            name,
        ):
            split_strategy = []
            # The first part of the split:
            # the dictionary means at dimension 0, extract the part of offset 0 and length 1 when regarding the whole 0 dimension as length 3.
            split_strategy.append(
                (
                    name.replace("qkv", "q"),
                    {0: {"offset": 0, "total_size": 3, "length": 1}},
                )
            )
            # The second part of the split:
            # the dictionary means at dimension 0, extract the part of offset 1 and length 1 when regarding the whole 0 dimension as length 3.
            split_strategy.append(
                (
                    name.replace("qkv", "k"),
                    {0: {"offset": 1, "total_size": 3, "length": 1}},
                )
            )
            # The third part of the split:
            # the dictionary means at dimension 0, extract the part of offset 2 and length 1 when regarding the whole 0 dimension as length 3.
            split_strategy.append(
                (
                    name.replace("qkv", "v"),
                    {0: {"offset": 2, "total_size": 3, "length": 1}},
                )
            )
            return split_strategy
        return []

    def get_unsplited_weight_name(self, weight_key: str) -> str:
        for key in ["q_proj", "k_proj", "v_proj"]:
            if key in weight_key:
                return weight_key.replace(key, "qkv_proj")
        for key in ["gate_proj", "up_proj"]:
            if key in weight_key:
                return weight_key.replace(key, "gate_up_proj")
        for key in ["q", "k", "v"]:
            if "visual" in weight_key and key in weight_key:
                return weight_key.replace(key, "qkv")
        return weight_key  # return full weight key



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import torch
import torch.multiprocessing as mp
from cosmos_rl.utils.pynccl import (
    create_nccl_uid,
    create_nccl_comm,
    nccl_send,
    nccl_recv,
    nccl_broadcast,
    nccl_allreduce,
)
from cosmos_rl.utils.pynccl_wrapper import ncclRedOpTypeEnum


def setup_nccl_comm(rank, world_size, nccl_uid):
    """Setup NCCL communicator for a process."""
    # Set device for this process (now using the visible device)
    torch.cuda.set_device(rank)  # Always use first visible device
    # Create NCCL communicator
    comm_idx = create_nccl_comm(nccl_uid, rank, world_size)
    return comm_idx


class TestNCCLBidirectionalSendRecv(unittest.TestCase):
    @staticmethod
    def run_bidirectional_sender(rank, world_size, nccl_uid, dtypes):
        """Run sender part of bidirectional NCCL send/recv test."""
        comm_idx = setup_nccl_comm(rank, world_size, nccl_uid)

        for dtype in dtypes:
            # Create test tensor
            tensor_size = 1000
            send_tensor = torch.ones(
                tensor_size, dtype=dtype, device=f"cuda:{rank}"
            ) * (rank + 1)
            recv_tensor = torch.zeros(tensor_size, dtype=dtype, device=f"cuda:{rank}")

            send_rank = 0
            # Send to other rank and receive from other rank
            other_rank = 1 - rank
            if rank == send_rank:
                nccl_send(send_tensor, other_rank, comm_idx)
                nccl_recv(recv_tensor, other_rank, comm_idx)
            else:
                nccl_recv(recv_tensor, other_rank, comm_idx)
                nccl_send(recv_tensor, other_rank, comm_idx)

            # Verify received data
            expected = torch.ones(tensor_size, dtype=dtype, device=f"cuda:{rank}") * (
                send_rank + 1
            )
            assert torch.allclose(recv_tensor, expected)

    def test_nccl_bidirectional_send_recv(self):
        """Test bidirectional NCCL send/recv operations between two processes with different CUDA devices."""
        world_size = 2

        # Create NCCL unique ID
        nccl_uid = create_nccl_uid()

        # Define functions for each process (same function but different rank)
        functions = self.run_bidirectional_sender
        # Spawn processes with different functions
        dtypes = [
            torch.float32,
            torch.float16,
            torch.int32,
            torch.int64,
            torch.uint8,
            torch.int8,
            torch.bfloat16,
            torch.float64,
        ]
        mp.spawn(
            functions,
            args=(world_size, nccl_uid, dtypes),
            nprocs=world_size,
            join=True,
        )


class TestNCCLBroadcast(unittest.TestCase):
    @staticmethod
    def run_broadcast(rank, world_size, nccl_uid, dtypes):
        """Run broadcast test for different data types."""
        comm_idx = setup_nccl_comm(rank, world_size, nccl_uid)

        for dtype in dtypes:
            # Test broadcasting from each rank
            for root_rank in range(world_size):
                # Create test tensor
                tensor_size = 1000
                if rank == root_rank:  # Root rank
                    # Create tensor with unique values based on root rank
                    tensor = torch.arange(
                        tensor_size, dtype=dtype, device=f"cuda:{rank}"
                    ) * (root_rank + 1)
                else:
                    # Create empty tensor for receiving
                    tensor = torch.zeros(
                        tensor_size, dtype=dtype, device=f"cuda:{rank}"
                    )

                # Perform broadcast from current root rank
                nccl_broadcast(tensor, root_rank, comm_idx)

                # Verify received data
                expected = torch.arange(
                    tensor_size, dtype=dtype, device=f"cuda:{rank}"
                ) * (root_rank + 1)
                assert torch.allclose(
                    tensor, expected
                ), f"Broadcast from rank {root_rank} failed for dtype {dtype}"

    def test_nccl_broadcast(self):
        """Test NCCL broadcast operations between multiple processes with different CUDA devices."""
        world_size = 4

        # Create NCCL unique ID
        nccl_uid = create_nccl_uid()

        # Define data types to test
        dtypes = [
            torch.float32,
            torch.float16,
            torch.int32,
            torch.int64,
            torch.uint8,
            torch.int8,
            torch.bfloat16,
            torch.float64,
        ]

        # Spawn processes
        mp.spawn(
            self.run_broadcast,
            args=(world_size, nccl_uid, dtypes),
            nprocs=world_size,
            join=True,
        )


class TestNCCLAllreduce(unittest.TestCase):
    @staticmethod
    def run_allreduce(rank, world_size, nccl_uid, dtypes):
        """Run allreduce test for different data types."""
        comm_idx = setup_nccl_comm(rank, world_size, nccl_uid)

        for dtype in dtypes:
            # Create test tensor
            tensor_size = 1000
            # Each rank creates a tensor with its rank value
            tensor = torch.ones(tensor_size, dtype=dtype, device=f"cuda:{rank}") * (
                rank + 1
            )
            op = ncclRedOpTypeEnum.from_torch(torch.distributed.ReduceOp.SUM)
            # Perform allreduce (sum)
            nccl_allreduce(tensor, tensor, op, comm_idx)

            # Verify result
            # For 4 ranks, the sum should be 1 + 2 + 3 + 4 = 10
            expected_sum = (
                torch.ones(tensor_size, dtype=dtype, device=f"cuda:{rank}") * 10
            )
            assert torch.allclose(tensor, expected_sum)

    def test_nccl_allreduce(self):
        """Test NCCL allreduce operations between multiple processes with different CUDA devices."""
        world_size = 4

        # Create NCCL unique ID
        nccl_uid = create_nccl_uid()

        # Define data types to test
        dtypes = [
            torch.float32,
            torch.float16,
            torch.int32,
            torch.int64,
            torch.uint8,
            torch.int8,
            torch.bfloat16,
            torch.float64,
        ]

        # Spawn processes
        mp.spawn(
            self.run_allreduce,
            args=(world_size, nccl_uid, dtypes),
            nprocs=world_size,
            join=True,
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from cosmos_rl.utils.logging import logger
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.policy.config import Config as RolloutConfig
from cosmos_rl.utils.distributed import (
    init_distributed,
    destroy_distributed,
    get_controller_metadata,
)
from cosmos_rl.rollout.vllm_rollout.vllm_rollout_worker import vLLMRolloutWorker


def run_rollout(*args, **kwargs):
    ctrl_ip, ctrl_port, metadata = get_controller_metadata()

    if metadata["config"] is None:
        raise RuntimeError(
            f"[Rollout] Please first go to http://{ctrl_ip}:{ctrl_port} to configure training parameters."
        )

    cosmos_rollout_config = RolloutConfig.from_dict(
        metadata["config"]
    )  # just use config as key temporarily

    task_type = cosmos_rollout_config.train.train_policy.type
    if task_type not in ["grpo"]:
        logger.info(
            "[Rollout] Task in controller is not type of Reinforcement Learning. Aborted."
        )
        sys.exit(0)

    logger.info(
        f"[Rollout] Loaded rollout configuration: {cosmos_rollout_config.rollout.model_dump()}"
    )

    try:
        rollout_backend = cosmos_rollout_config.rollout.backend
        if rollout_backend == "vllm":
            parallel_dims = ParallelDims.from_config(
                parallesim_config=cosmos_rollout_config.rollout.parallelism
            )
            init_distributed()
            parallel_dims.build_mesh(device_type="cuda")
            rollout_worker = vLLMRolloutWorker(cosmos_rollout_config, parallel_dims)

        elif rollout_backend == "trtllm":
            try:
                from cosmos_rl.rollout.trtllm_rollout.trtllm_rollout_wrapper import (
                    TRTLLMRolloutWrapper,
                )
            except ImportError as e:
                logger.error(f"[Rollout] TRTLLMRolloutWrapper importing failed! {e}")
                raise e
            # if backend is trtllm, we leave distribution initialization to trtllm executor.
            rollout_worker = TRTLLMRolloutWrapper(cosmos_rollout_config)
        else:
            raise ValueError(f"Invalid rollout backend: {rollout_backend}")
        rollout_worker.work()
    except Exception:
        import traceback

        traceback.print_exc()
    finally:
        destroy_distributed()
        logger.info("[Rollout] Destroy context of torch dist.")


if __name__ == "__main__":
    run_rollout()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List

import torch
from transformers import AutoTokenizer

from cosmos_rl.dispatcher.data.packer.decoder_only_llm_data_packer import (
    DecoderOnlyLLMDataPacker,
)
from cosmos_rl.policy.config import Config


class DeepSeek_DataPacker(DecoderOnlyLLMDataPacker):
    """
    Data protocol & processing logic for the decoder only LLM for SFT and RL training.
    """

    def setup(self, config: Config, tokenizer: AutoTokenizer, *args, **kwargs):
        super().setup(config, tokenizer, *args, **kwargs)
        self.seq_len = config.policy.model_max_length

    def policy_compute_max_len(
        self, processed_samples: List[DecoderOnlyLLMDataPacker.RLPolicyInput]
    ) -> int:
        return max(self.seq_len, max([len(x.input_ids) for x in processed_samples]))

    def policy_collate_fn(
        self,
        processed_samples: List[DecoderOnlyLLMDataPacker.RLPolicyInput],
        computed_max_len: int,
    ) -> Dict[str, Any]:
        computed_max_len = max(computed_max_len, self.seq_len)
        input_ids = [x.input_ids for x in processed_samples]
        logprob_masks = [x.logprob_masks for x in processed_samples]
        assert len(input_ids) == len(
            logprob_masks
        ), "The length of input_ids, and logprob_masks should be the same"
        device = torch.cuda.current_device()

        collated_dict = {}
        collated_dict["input_ids"] = torch.tensor(
            [
                x[:computed_max_len]
                + [self.tokenizer.pad_token_id] * (max(0, computed_max_len - len(x)))
                for x in input_ids
            ],
            dtype=torch.long,
        ).to(device)
        collated_dict["logprob_masks"] = torch.tensor(
            [
                x[:computed_max_len] + [0] * (max(0, computed_max_len - len(x)))
                for x in logprob_masks
            ],
            dtype=torch.bool,
        ).to(device)

        return collated_dict

    def sft_collate_fn(
        self,
        processed_samples: List[Dict[str, Any]],
        computed_max_len: int,
        pad_token_id: int,
        ignore_label_id: int,
    ) -> Dict[str, Any]:
        """
        Collate the processed samples into a minibatch dictionary passed to the SFT model.
        """
        computed_max_len = max(computed_max_len, self.seq_len)
        # First truncate the samples to the computed_max_len
        list_of_input_ids = [
            x["token_ids"][:computed_max_len] for x in processed_samples
        ]
        list_of_label_ids = [
            x["label_ids"][:computed_max_len] for x in processed_samples
        ]

        # Then pad the samples to the computed_max_len
        input_ids = torch.tensor(
            [
                x[:computed_max_len]
                + [pad_token_id] * (max(0, computed_max_len - len(x)))
                for x in list_of_input_ids
            ],
            dtype=torch.long,
        )
        # Model accept unshifted label_ids for loss computation
        label_ids = torch.tensor(
            [
                x[:computed_max_len]
                + [ignore_label_id] * (max(0, computed_max_len - len(x)))
                for x in list_of_label_ids
            ],
            dtype=torch.long,
        )

        return {
            "input_ids": input_ids,
            "label_ids": label_ids,
        }



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest
import torch
import subprocess
import sys
from multiprocessing import shared_memory
import numpy as np
from cosmos_rl.utils.pynccl import (
    create_nccl_uid,
)


class TestPolicyToPolicy(unittest.TestCase):
    def policy_to_policy_unicast(self):
        """Test NCCL communication between multiple ranks using torchrun."""
        cur_dir = os.path.dirname(os.path.abspath(__file__))

        # Create NCCL UID and shared memory
        nccl_uid = create_nccl_uid()
        nccl_uid_tensor = torch.tensor(nccl_uid, dtype=torch.int64)
        shms = []
        world_size = 2
        for i in range(world_size):
            shms.append(
                shared_memory.SharedMemory(
                    create=True,
                    size=(nccl_uid_tensor.numel() + 1) * nccl_uid_tensor.element_size(),
                )
            )
            uid_array = np.ndarray(
                (nccl_uid_tensor.numel() + 1,), dtype=np.int64, buffer=shms[-1].buf
            )
            uid_array[-1] = 0

        try:
            # Create the Python command for torchrun
            policy_cmd = [
                "torchrun",
                f"--nproc_per_node={world_size}",  # Use 2 GPUs
                "--role=rank",
                "--tee=3",
                "--rdzv_backend=c10d",
                "--rdzv_endpoint=localhost:0",
                os.path.join(cur_dir, "launch_test_worker.py"),
                "--shm_name",
                ",".join([shm.name for shm in shms]),
                "--shm_size",
                str(nccl_uid_tensor.numel()),
                "--mode",
                "policy_send_to_policy",
            ]
            policy_dst_cmd = [
                "torchrun",
                f"--nproc_per_node={world_size}",  # Use 2 GPUs
                "--role=rank",
                "--tee=3",
                "--rdzv_backend=c10d",
                "--rdzv_endpoint=localhost:0",
                os.path.join(cur_dir, "launch_test_worker.py"),
                "--shm_name",
                ",".join([shm.name for shm in shms]),
                "--shm_size",
                str(nccl_uid_tensor.numel()),
                "--mode",
                "policy_recv_from_policy",
            ]
            policy_env = dict(os.environ)
            policy_env["CUDA_VISIBLE_DEVICES"] = "0,1"
            # Start the process
            policy_process = subprocess.Popen(
                policy_cmd,
                stdout=sys.stderr,
                stderr=sys.stderr,
                env=policy_env,
            )
            policy_dst_env = dict(os.environ)
            policy_dst_env["CUDA_VISIBLE_DEVICES"] = "2,3"
            policy_dst_process = subprocess.Popen(
                policy_dst_cmd,
                stdout=sys.stderr,
                stderr=sys.stderr,
                env=policy_dst_env,
            )

            try:
                # Wait for process to complete
                for process in [policy_process, policy_dst_process]:
                    stdout, stderr = process.communicate()

                    # Check if process completed successfully
                    assert process.returncode == 0, f"Process failed: {stderr.decode()}"

            finally:
                # Ensure process is terminated
                for process in [policy_process, policy_dst_process]:
                    process.wait()
        finally:
            # Clean up shared memory
            try:
                for shm in shms:
                    shm.close()
                    shm.unlink()
            except FileNotFoundError:
                # Ignore if shared memory is already unlinked
                pass

    def policy_to_policy_broadcast(self):
        """Test NCCL communication between multiple ranks using torchrun."""
        cur_dir = os.path.dirname(os.path.abspath(__file__))

        # Create NCCL UID and shared memory
        nccl_uid = create_nccl_uid()
        nccl_uid_tensor = torch.tensor(nccl_uid, dtype=torch.int64)
        shms = []
        world_size = 2
        for i in range(world_size):
            shms.append(
                shared_memory.SharedMemory(
                    create=True,
                    size=(nccl_uid_tensor.numel() + 1) * nccl_uid_tensor.element_size(),
                )
            )
            uid_array = np.ndarray(
                (nccl_uid_tensor.numel() + 1,), dtype=np.int64, buffer=shms[-1].buf
            )
            uid_array[-1] = 0

        try:
            # Create the Python command for torchrun
            policy_cmd = [
                "torchrun",
                f"--nproc_per_node={world_size}",  # Use 2 GPUs
                "--role=rank",
                "--tee=3",
                "--rdzv_backend=c10d",
                "--rdzv_endpoint=localhost:0",
                os.path.join(cur_dir, "launch_test_worker.py"),
                "--shm_name",
                ",".join([shm.name for shm in shms]),
                "--shm_size",
                str(nccl_uid_tensor.numel()),
                "--mode",
                ",".join(["policy_broadcast_to_policy", "4", "0"]),
            ]
            policy_dst_cmd = [
                "torchrun",
                f"--nproc_per_node={world_size}",  # Use 2 GPUs
                "--role=rank",
                "--tee=3",
                "--rdzv_backend=c10d",
                "--rdzv_endpoint=localhost:0",
                os.path.join(cur_dir, "launch_test_worker.py"),
                "--shm_name",
                ",".join([shm.name for shm in shms]),
                "--shm_size",
                str(nccl_uid_tensor.numel()),
                "--mode",
                ",".join(["policy_broadcast_to_policy", "4", "0"]),
            ]
            policy_env = dict(os.environ)
            policy_env["CUDA_VISIBLE_DEVICES"] = "0,1"
            # Start the process
            policy_process = subprocess.Popen(
                policy_cmd,
                stdout=sys.stderr,
                stderr=sys.stderr,
                env=policy_env,
            )
            policy_dst_env = dict(os.environ)
            policy_dst_process = []

            envs = ["2,3", "4,5", "6,7"]
            reps = ["1", "2", "3"]
            for env, rep in zip(envs, reps):
                policy_dst_env["CUDA_VISIBLE_DEVICES"] = env
                policy_dst_cmd[-1] = ",".join(["policy_broadcast_to_policy", "4", rep])
                policy_dst_process.append(
                    subprocess.Popen(
                        policy_dst_cmd,
                        stdout=sys.stderr,
                        stderr=sys.stderr,
                        env=policy_dst_env,
                    )
                )
            try:
                # Wait for process to complete
                for process in [policy_process] + policy_dst_process:
                    stdout, stderr = process.communicate()

                    # Check if process completed successfully
                    assert process.returncode == 0, f"Process failed: {stderr.decode()}"

            finally:
                # Ensure process is terminated
                for process in [policy_process] + policy_dst_process:
                    process.wait()
        finally:
            # Clean up shared memory
            try:
                for shm in shms:
                    shm.close()
                    shm.unlink()
            except FileNotFoundError:
                # Ignore if shared memory is already unlinked
                pass

    def test_policy_to_policy_unicast_all_params(self):
        self.policy_to_policy_unicast()

    def test_policy_to_policy_broadcast_all_params(self):
        self.policy_to_policy_broadcast()


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import torch
from torch import nn
from dataclasses import dataclass, field
from typing import Tuple, List, Optional, Callable
from transformers import AutoConfig
from cosmos_rl.utils.util import (
    resolve_model_path,
    IdentityLayer,
    clear_weight_name,
    sync_model_vocab,
    retry,
)
from safetensors import safe_open
from cosmos_rl.policy.model.base import ModelRegistry, BaseModel
from cosmos_rl.policy.model.gpt.weight_mapper import GPTWeightMapper
from cosmos_rl.utils.logging import logger
from cosmos_rl.policy.model.gpt.weight_converter import convert_weight_from_hf
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.policy.config import Config as CosmosConfig
from transformers.modeling_rope_utils import ROPE_INIT_FUNCTIONS
from functools import cached_property
import cosmos_rl.policy.kernel.modeling_utils as modeling_utils
from cosmos_rl.policy.kernel.norm import RMSNorm
import cosmos_rl.policy.kernel.rope as rope
from cosmos_rl.policy.kernel.fused import MLPActMulFunc
from cosmos_rl.utils.sequence_packing import pack_sequences_for_inputs


def build_norm(
    norm_type: str, dim: int, eps: float, casting_mode: Optional[str] = None
):
    assert norm_type == "rmsnorm", f"Unknown norm_type: '{norm_type}'"
    return RMSNorm(dim, eps, casting_mode=casting_mode)


@dataclass
class GPTArgs:
    dim: int
    ffn_dim: int
    n_layers: int
    n_heads: int
    n_kv_heads: int
    head_dim: int
    vocab_size: int
    max_seq_len: int
    biases: List[str] = field(default_factory=lambda: [])
    q_k_norm_enabled: bool = False
    norm_eps: float = 1e-6
    rope_theta: float = 10000
    norm_type: str = "rmsnorm"
    rope_type: str = "default"
    hf_config: AutoConfig = None


class RotaryEmbedding(nn.Module):
    def __init__(self, args: GPTArgs, device=None):
        super().__init__()
        self.args = args
        self.rope_init_fn = ROPE_INIT_FUNCTIONS[args.rope_type]
        self.device = device
        self.config = args
        self.reset_inv_freq(device=device)

    def reset_inv_freq(self, device: torch.device = None):
        inv_freq, self.attention_scaling = self.rope_init_fn(
            self.config.hf_config, self.device
        )
        if not hasattr(self, "inv_freq"):
            self.register_buffer("inv_freq", inv_freq, persistent=False)
        else:
            self.inv_freq.to(torch.float32)
            with torch.no_grad():
                self.inv_freq.data.copy_(inv_freq)

    @torch.no_grad()
    def forward(self, x, position_ids):
        if self.inv_freq.dtype != torch.float32:
            self.reset_inv_freq(device=x.device)
            assert self.inv_freq.dtype == torch.float32

        inv_freq_expanded = (
            self.inv_freq[None, :, None].float().expand(position_ids.shape[0], -1, 1)
        )
        position_ids_expanded = position_ids[:, None, :].float()
        # Force float32 (see https://github.com/huggingface/transformers/pull/29285)
        device_type = x.device.type
        device_type = (
            device_type
            if isinstance(device_type, str) and device_type != "mps"
            else "cpu"
        )
        with torch.autocast(device_type=device_type, enabled=False):
            freqs = (
                inv_freq_expanded.float().to(x.device) @ position_ids_expanded.float()
            ).transpose(1, 2)
            emb = torch.cat((freqs, freqs), dim=-1)
            cos = emb.cos()
            sin = emb.sin()

        # Advanced RoPE types (e.g. yarn) apply a post-processing scaling factor, equivalent to scaling attention
        cos = cos * self.attention_scaling
        sin = sin * self.attention_scaling

        return cos.to(dtype=x.dtype), sin.to(dtype=x.dtype)


class Attention(nn.Module):
    """
    Multi-head attention module.

    Args:
        model_args (GPTArgs): Model configuration arguments.

    Attributes:
        n_kv_heads (int): Number of key and value heads.
        n_heads (int): Number of query heads.
        n_rep (int): Number of repetitions for local heads.
        head_dim (int): Dimension size of each attention head.
        q_proj (Linear): Linear transformation for queries.
        k_proj (Linear): Linear transformation for keys.
        v_proj (Linear): Linear transformation for values.
        o_proj (Linear): Linear transformation for output.
    """

    def __init__(self, model_args: GPTArgs):
        super().__init__()
        self.n_heads = model_args.n_heads
        self.n_kv_heads = model_args.n_kv_heads
        self.n_rep = self.n_heads // self.n_kv_heads
        self.head_dim = model_args.head_dim
        self.attn_func = modeling_utils.flash_attn_func
        self.attn_func_varlen = modeling_utils.flash_attn_varlen_func

        self.q_proj = nn.Linear(
            model_args.dim,
            model_args.n_heads * self.head_dim,
            bias="q_proj" in model_args.biases,
        )
        self.q_norm = (
            build_norm(
                model_args.norm_type,
                dim=self.head_dim,
                eps=model_args.norm_eps,
                casting_mode=model_args.hf_config.model_type,
            )
            if model_args.q_k_norm_enabled
            else None
        )

        self.k_proj = nn.Linear(
            model_args.dim,
            self.n_kv_heads * self.head_dim,
            bias="k_proj" in model_args.biases,
        )
        self.k_norm = (
            build_norm(
                model_args.norm_type,
                dim=self.head_dim,
                eps=model_args.norm_eps,
                casting_mode=model_args.hf_config.model_type,
            )
            if model_args.q_k_norm_enabled
            else None
        )

        self.v_proj = nn.Linear(
            model_args.dim,
            self.n_kv_heads * self.head_dim,
            bias="v_proj" in model_args.biases,
        )
        self.o_proj = nn.Linear(
            model_args.n_heads * self.head_dim,
            model_args.dim,
            bias="o_proj" in model_args.biases,
        )
        self.rope_func = rope.RotaryPositionEmbedding()

    def forward(
        self,
        x: torch.Tensor,
        position_embeddings: Tuple[torch.Tensor],
        cu_seqlens: Optional[torch.Tensor] = None,
        max_seqlen: Optional[int] = None,
    ):
        """
        Forward pass of the attention module.

        Args:
            x (torch.Tensor): Input tensor.
            position_embeddings (torch.Tensor): Position embeddings.
            cu_seqlens (torch.Tensor, optional): Cumulative sequence lengths for variable-length sequences.
            max_seqlen (int, optional): Maximum sequence length for variable-length sequences.

        Returns:
            torch.Tensor: Output tensor after attention.

        """

        bs, seqlen, _ = x.shape
        xq, xk, xv = self.q_proj(x), self.k_proj(x), self.v_proj(x)
        if self.q_norm is not None:
            xq = self.q_norm(xq.view(bs, seqlen, -1, self.head_dim))
        if self.k_norm is not None:
            xk = self.k_norm(xk.view(bs, seqlen, -1, self.head_dim))

        # Use -1 instead of `n_heads` (or `n_kv_heads`) to infer the actual
        # local heads from sizes of xq, xk, and xv as TP may have sharded them
        # after the above linear ops.
        xq = xq.view(bs, seqlen, -1, self.head_dim)
        xk = xk.view(bs, seqlen, -1, self.head_dim)
        xv = xv.view(bs, seqlen, -1, self.head_dim)

        cos, sin = position_embeddings
        xq, xk = self.rope_func(xq, xk, cos, sin, unsqueeze_dim=2)

        input_dtype = xq.dtype
        if input_dtype == torch.float32:
            if torch.is_autocast_enabled():
                target_dtype = torch.get_autocast_gpu_dtype()
            else:
                raise ValueError("Flash attention only supports float32 input")
            xq = xq.to(target_dtype)
            xk = xk.to(target_dtype)
            xv = xv.to(target_dtype)

        if cu_seqlens is not None:
            assert (
                max_seqlen is not None
            ), "max_seqlen must be provided for variable-length sequences"
            xq = xq.view(seqlen, -1, self.head_dim)
            xk = xk.view(seqlen, -1, self.head_dim)
            xv = xv.view(seqlen, -1, self.head_dim)
            output = self.attn_func_varlen(
                xq,
                xk,
                xv,
                cu_seqlens,
                cu_seqlens,
                max_seqlen,
                max_seqlen,
                causal=True,
            )
        else:
            output = self.attn_func(xq, xk, xv, causal=True)
        output = output.view(bs, seqlen, -1)
        return self.o_proj(output)


class FeedForward(nn.Module):
    """
    FeedForward module

    Args:
        dim (int): Input dimension.
        hidden_dim (int): Hidden dimension of the feedforward layer.

    Attributes:
        w1 (Linear): Linear transformation for the first layer.
        w2 (Linear): Linear transformation for the second layer.
        w3 (Linear): Linear transformation for the third layer.

    """

    def __init__(
        self,
        dim: int,
        hidden_dim: int,
        model_args: GPTArgs,
    ):
        super().__init__()
        self.up_proj = nn.Linear(dim, hidden_dim, bias="up_proj" in model_args.biases)
        self.down_proj = nn.Linear(
            hidden_dim, dim, bias="down_proj" in model_args.biases
        )
        self.gate_proj = nn.Linear(
            dim, hidden_dim, bias="gate_proj" in model_args.biases
        )
        self.act_mul_func = MLPActMulFunc(nn.SiLU())

    def forward(self, x):
        return self.down_proj(self.act_mul_func(self.gate_proj(x), self.up_proj(x)))


class GPTBlock(nn.Module):
    """
    GPTBlock Module

    Args:
        layer_id (int): Identifier for the layer.
        model_args (GPTArgs): Model configuration arguments.

    Attributes:
        n_heads (int): Number of attention heads.
        dim (int): Dimension size of the model.
        head_dim (int): Dimension size of each attention head.
        attention (Attention): Attention module.
        feed_forward (FeedForward): FeedForward module.
        layer_id (int): Identifier for the layer.
        attention_norm (RMSNorm): Layer normalization for attention output.
        ffn_norm (RMSNorm): Layer normalization for feedforward output.

    """

    def __init__(self, layer_id: int, model_args: GPTArgs):
        super().__init__()
        self.n_heads = model_args.n_heads
        self.dim = model_args.dim
        self.self_attn = Attention(model_args)
        self.mlp = FeedForward(
            dim=model_args.dim,
            hidden_dim=model_args.ffn_dim,
            model_args=model_args,
        )
        self.layer_id = layer_id
        self.num_layers = model_args.n_layers

        self.input_layernorm = build_norm(
            model_args.norm_type,
            dim=model_args.dim,
            eps=model_args.norm_eps,
            casting_mode=model_args.hf_config.model_type,
        )
        self.post_attention_layernorm = build_norm(
            model_args.norm_type,
            dim=model_args.dim,
            eps=model_args.norm_eps,
            casting_mode=model_args.hf_config.model_type,
        )

    def forward(
        self,
        x: torch.Tensor,
        position_embeddings: Optional[
            Tuple[torch.Tensor, torch.Tensor]
        ] = None,  # necessary, but kept here for BC
        **kwargs,
    ):
        """
        Perform a forward pass through the GPTBlock.

        Args:
            x (torch.Tensor): Input tensor.
            position_embeddings (torch.Tensor): Position embeddings.
            kwargs: Additional keyword arguments.

        Returns:
            torch.Tensor: Output tensor after applying attention and feedforward layers.

        """
        h = x + self.self_attn(
            self.input_layernorm(x),
            position_embeddings,
            cu_seqlens=kwargs.get("cu_seqlens", None),
            max_seqlen=kwargs.get("max_seqlen", None),
        )
        out = h + self.mlp(self.post_attention_layernorm(h))
        return out


@ModelRegistry.register(GPTWeightMapper)
class GPT(BaseModel):
    """
    GPT Module

    Args:
        model_args (GPTArgs): Model configuration arguments.

    Attributes:
        model_args (GPTArgs): Model configuration arguments.
        vocab_size (int): Vocabulary size.
        n_layers (int): Number of layers in the model.
        embed_tokens (ParallelEmbedding): Token embeddings.
        layers (torch.nn.ModuleList): List of GPT blocks.
        norm (RMSNorm): Layer normalization for the model output.
        lm_head (ColumnParallelLinear): Linear layer for final output.
    """

    @staticmethod
    def supported_model_types():
        return ["llama", "qwen2", "qwen3"]

    def __init__(self, model_args: GPTArgs):
        super().__init__(model_args.hf_config)

        self.model_args = model_args
        self.vocab_size = model_args.vocab_size
        self.n_layers = model_args.n_layers
        self.rotary_emb = RotaryEmbedding(model_args)

        self.embed_tokens = nn.Embedding(model_args.vocab_size, model_args.dim)

        self.layers = torch.nn.ModuleDict()
        for layer_id in range(model_args.n_layers):
            self.layers[str(layer_id)] = GPTBlock(layer_id, model_args)

        self.norm = build_norm(
            model_args.norm_type,
            dim=model_args.dim,
            eps=model_args.norm_eps,
            casting_mode=model_args.hf_config.model_type,
        )

        if not model_args.hf_config.tie_word_embeddings:
            self.tie_embed_tokens = False
            self.lm_head = nn.Linear(
                model_args.dim,
                model_args.vocab_size,
                bias="lm_head" in model_args.biases,
            )
        else:
            self.tie_embed_tokens = True
        self.identity_layer = IdentityLayer()

    def forward(
        self,
        input_ids: torch.Tensor,
        position_ids: torch.Tensor = None,
        interested_tokens: Optional[torch.BoolTensor] = None,
        *args,
        **kwargs,
    ):
        if self.embed_tokens is not None:
            inputs_embeds = self.embed_tokens(input_ids)
            # Do not remove this line
            # This is a trick for TP with torch.compile
            h = self.identity_layer(inputs_embeds)
        else:
            inputs_embeds = input_ids
            h = inputs_embeds

        position_embeddings = self.rotary_emb(h, position_ids.to(dtype=torch.long))

        if "valid_input_len" in kwargs:
            valid_input_len = kwargs["valid_input_len"]
            updated_kwargs = pack_sequences_for_inputs(
                inputs_embeds,
                valid_input_len,
                list(position_embeddings),
                interested_tokens,
                inputs_seq_dim=1,
                inputs_batch_dim=0,
                position_ids_seq_dim=1,
                position_ids_batch_dim=0,
                interested_tokens_seq_dim=1,
                interested_tokens_batch_dim=0,
                padding_mask=kwargs.get("padding_mask", None),
                cp_mesh=kwargs.get("cp_mesh", None),
            )
            position_embeddings = tuple(updated_kwargs.pop("position_ids"))
            interested_tokens = updated_kwargs.pop("interested_tokens")
            h = updated_kwargs.pop("inputs")
            h = self.identity_layer(h)
            kwargs.update(updated_kwargs)

        for layer in self.layers.values():
            if (
                hasattr(layer, "_gradient_checkpointing_enabled")
                and layer._gradient_checkpointing_enabled
            ):
                h = torch.utils.checkpoint.checkpoint(
                    layer,
                    h,
                    position_embeddings,
                    **kwargs,
                    use_reentrant=False,
                )
            else:
                h = layer(h, position_embeddings=position_embeddings, **kwargs)

        # Add `if` check just in case `pp` is enabled
        if self.norm is not None:
            if interested_tokens is not None:
                assert not isinstance(
                    h, torch.distributed.tensor.DTensor
                ), "logprob_masks must be a local tensor"
                h = h[interested_tokens]
            h = self.norm(h)
            if not self.tie_embed_tokens:
                output = self.lm_head(h)
            else:
                is_w_dist_tensor = isinstance(
                    self.embed_tokens.weight, torch.distributed.tensor.DTensor
                )
                embed_tokens_weight = (
                    self.embed_tokens.weight.full_tensor()
                    if is_w_dist_tensor
                    else self.embed_tokens.weight
                )
                is_a_dist_tensor = isinstance(h, torch.distributed.tensor.DTensor)
                h = h.full_tensor() if is_a_dist_tensor else h
                output = h @ embed_tokens_weight.t()
            return output
        else:
            return h

    def post_to_empty_hook(self, cosmos_config: CosmosConfig):
        # rotary.inv_freq could get deleted and not re-initialized
        # so we need to delete it manually
        self.rotary_emb.to(torch.cuda.current_device())
        self.rotary_emb.reset_inv_freq()

    @property
    def parallelize_fn(self):
        from cosmos_rl.policy.model.gpt.parallelize import parallelize

        return parallelize, self

    def apply_pipeline_split(self, pp_rank, pp_size):
        """
        Apply pipeline split to the model.
        This typically involves splitting the model into multiple stages,
        and moving each stage to a different device.
        """
        assert pp_size > 1
        is_first = pp_rank == 0
        is_last = pp_rank == pp_size - 1

        # Compute the layers belonging to this stage
        n_layers = len(self.layers)
        layers_per_stage = n_layers // pp_size

        if not is_first:
            self.embed_tokens = None
        if not is_last:
            self.lm_head = None
            self.norm = None

        local_layers = torch.nn.ModuleDict()
        for i in range(
            pp_rank * layers_per_stage,
            ((pp_rank + 1) * layers_per_stage) if not is_last else n_layers,
        ):
            local_layers[str(i)] = self.layers[str(i)]

        # Reset the layers for pipeline splitting
        self.layers = local_layers

    def load_hf_weights(
        self,
        model_name_or_path: str,
        parallel_dims: ParallelDims,
        device: torch.device,
        revision: Optional[str] = None,
    ):
        """
        Load weights from a HuggingFace model.

        Args:
            model_path (str): Path to the HuggingFace model.
            parallel_dims (ParallelDims): Parallel dimensions definition.
            info_inly (bool): Only collect the tensor infomation without actual data loading.
        """
        # Load all safetensors from `model_path`
        model_type = retry(AutoConfig.from_pretrained)(model_name_or_path).model_type
        model_path = resolve_model_path(model_name_or_path, revision=revision)
        safetensors_files = [
            f for f in os.listdir(model_path) if f.endswith(".safetensors")
        ]

        self_state_dict = self.state_dict()
        self_state_dict = {clear_weight_name(k): v for k, v in self_state_dict.items()}
        lm_head_weight_key = "lm_head.weight"
        embed_tokens_weight_key = "model.embed_tokens.weight"
        weights_of_ckpt_names = set()
        reserved = {}
        for f in safetensors_files:
            weights_of_ckpt = {}
            ckpt = safe_open(
                os.path.join(model_path, f), framework="pt", device=str(device)
            )
            keys = ckpt.keys()
            for name in keys:
                ckpt_tensor = ckpt.get_tensor(name)
                weights_of_ckpt[name] = ckpt_tensor
                weights_of_ckpt_names.add(name)
                if name == embed_tokens_weight_key:
                    reserved[name] = ckpt_tensor

            for name in weights_of_ckpt.keys():
                tensor = weights_of_ckpt[name]
                dest_name, shared_weight = convert_weight_from_hf(
                    tensor, name, model_type, parallel_dims
                )
                if dest_name not in self_state_dict and parallel_dims.pp_enabled:
                    # logger.info(f"Weight `{dest_name}` is discarded, maybe due to pipeline parallelism. Skipping this weight checking")
                    continue

                target_tensor = self_state_dict[dest_name]
                is_dist_tensor = isinstance(
                    target_tensor, torch.distributed.tensor.DTensor
                )
                local_view = (
                    target_tensor.to_local() if is_dist_tensor else target_tensor
                )
                assert (
                    local_view.shape == shared_weight.shape
                ), f"Shape mismatch: {local_view.shape} != {shared_weight.shape} for {dest_name}"
                with torch.no_grad():
                    local_view.data.copy_(shared_weight)

        if (
            lm_head_weight_key not in weights_of_ckpt_names
            and embed_tokens_weight_key in weights_of_ckpt_names
        ):
            # tied with embed_tokens.weight
            name = lm_head_weight_key
            assert embed_tokens_weight_key in reserved
            tensor = reserved[embed_tokens_weight_key]
            dest_name, shared_weight = convert_weight_from_hf(
                tensor, name, model_type, parallel_dims
            )
            if dest_name in self_state_dict:
                target_tensor = self_state_dict[dest_name]
                is_dist_tensor = isinstance(
                    target_tensor, torch.distributed.tensor.DTensor
                )
                local_view = (
                    target_tensor.to_local() if is_dist_tensor else target_tensor
                )
                assert (
                    local_view.shape == shared_weight.shape
                ), f"Shape mismatch: {local_view.shape} != {shared_weight.shape} for {dest_name}"
                with torch.no_grad():
                    local_view.data.copy_(shared_weight)

    def get_position_ids(self, **kwargs) -> Tuple[torch.Tensor, torch.Tensor, int]:
        seq_dim_idx = 1
        inputs = kwargs["input_ids"]
        position_ids = (
            torch.arange(inputs.size(-1), dtype=torch.long, device=inputs.device)
            .unsqueeze(0)
            .expand_as(inputs)
        )
        return position_ids, inputs, seq_dim_idx

    def separate_model_parts(self) -> List[nn.Module]:
        return [self]

    @cached_property
    def _get_nparams_and_flops_fn(self) -> Callable[[int], tuple[int, int]]:
        nparams = sum(p.numel() for p in self.parameters())
        nparams_embedding = sum(
            sum(p.numel() for p in m.parameters())
            for m in self.children()
            if isinstance(m, nn.Embedding)
        )

        # Reasoning behind the factor of 12 for the self-attention part of the formula:
        # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
        # 2. the flash attention does 1 more matmul recomputation in the backward
        #    but recomputation should not be counted in calculating MFU           (+0)
        # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
        # 4. we follow the convention and do not account for sparsity in causal attention
        layers, heads, head_dim = (
            len(self.layers),
            self.model_args.n_heads,
            self.model_args.dim // self.model_args.n_heads,
        )
        return lambda seq_len: (
            nparams,
            6 * (nparams - nparams_embedding)
            + 12 * layers * heads * head_dim * seq_len,
        )

    def get_nparams_and_flops(self, seq_len: int) -> tuple[int, int]:
        return self._get_nparams_and_flops_fn(seq_len)

    @classmethod
    def from_model_args(cls, model_args: GPTArgs) -> "GPT":
        """
        Initialize a GPT model from a GPTArgs object.

        Args:
            model_args (GPTArgs): Model configuration arguments.

        Returns:
            GPT: GPT model.

        """
        return cls(model_args)

    @classmethod
    def from_pretrained(
        cls,
        hf_config: AutoConfig,
        model_name_or_path: str,
        max_position_embeddings: Optional[int] = None,
    ) -> "GPT":
        """
        Initialize a GPT model from a pretrained model.

        Args:
            model_name_or_path (str): Model name or path to the pretrained model.

        Returns:
            GPT: GPT model.

        """
        if hf_config.model_type not in cls.supported_model_types():
            raise ValueError(f"Unsupported model type: {hf_config.model_type}")

        if max_position_embeddings is None:
            max_position_embeddings = hf_config.max_position_embeddings
        else:
            hf_config.max_position_embeddings = max_position_embeddings

        vocab_size = sync_model_vocab(model_name_or_path)

        rope_scaling = {}
        if hasattr(hf_config, "rope_scaling"):
            rope_scaling = hf_config.rope_scaling or {}
        rope_type = rope_scaling.get("rope_type", "default")

        bias_list = {
            "qwen2": ["q_proj", "k_proj", "v_proj"],
            "llama": [],
            "qwen3": [],
        }[hf_config.model_type]

        try:
            head_dim = hf_config.head_dim
        except Exception:
            head_dim = hf_config.hidden_size // hf_config.num_attention_heads
            logger.warning(f"head_dim not found in config, using {head_dim}")

        model = cls.from_model_args(
            GPTArgs(
                dim=hf_config.hidden_size,
                ffn_dim=hf_config.intermediate_size,
                n_layers=hf_config.num_hidden_layers,
                n_heads=hf_config.num_attention_heads,
                n_kv_heads=hf_config.num_key_value_heads,
                head_dim=head_dim,
                vocab_size=vocab_size,
                max_seq_len=max_position_embeddings,
                rope_theta=hf_config.rope_theta,
                q_k_norm_enabled=hf_config.model_type == "qwen3",
                norm_type="rmsnorm",
                rope_type=rope_type,
                biases=bias_list,
                hf_config=hf_config,
            )
        )
        return model

    @classmethod
    def fqn_filter_for_fp8(cls) -> List[str]:
        return ["lm_head"]

    def check_cp_compatible(self, cp_size: int, tp_size: int):
        if not (self.model_args.n_heads % (cp_size * tp_size) == 0):
            raise ValueError(
                f"Model is not compatible with cp parallelism, model's head number={self.model_args.n_heads} is not divisible by cp size({cp_size}) * tp_size({tp_size}) = {cp_size * tp_size}"
            )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn as nn
import unittest

from cosmos_rl.policy.lora.plugin import (
    inject_lora_adapters,
    mark_only_lora_as_trainable,
    merge_lora_weights_,
    unmerge_lora_weights_,
    lora_state_dict,
    load_lora_state_dict,
    LoraInjectedLinear,
)
from cosmos_rl.policy.config import LoraConfig


def _base_state_dict(m):
    # drop any lora_* tensors; keep only base weights/biases
    return {k: v for k, v in m.state_dict().items() if "lora_" not in k}


class TinyBlock(nn.Module):
    def __init__(self, in_f=32, hidden=64, out_f=16):
        super().__init__()
        self.q_proj = nn.Linear(in_f, hidden)
        self.act = nn.GELU()
        self.v_proj = nn.Linear(hidden, out_f)

    def forward(self, x):
        return self.v_proj(self.act(self.q_proj(x)))


class LoRATest(unittest.TestCase):
    def test_injection_and_freeze(self):
        model = TinyBlock()
        cfg = LoraConfig(
            r=8, lora_alpha=16, lora_dropout=0.1, target_modules=["q_proj", "v_proj"]
        )
        model, replaced = inject_lora_adapters(model, cfg)

        # Both should be replaced
        assert any("q_proj" in n for n in replaced)
        assert any("v_proj" in n for n in replaced)
        assert isinstance(model.q_proj, LoraInjectedLinear)
        assert isinstance(model.v_proj, LoraInjectedLinear)

        # Freeze base, enable only LoRA params
        mark_only_lora_as_trainable(model, cfg)
        base_params = []
        lora_params = []
        for n, p in model.named_parameters():
            if "lora_" in n:
                lora_params.append((n, p))
            else:
                base_params.append((n, p))
        assert all(not p.requires_grad for _, p in base_params)
        assert all(p.requires_grad for _, p in lora_params)

    def test_forward_equivalence_merge_unmerge(self):
        x = torch.randn(4, 32)
        model = TinyBlock()
        cfg = LoraConfig(
            r=4, lora_alpha=8, lora_dropout=0.0, target_modules=["q_proj", "v_proj"]
        )
        model, _ = inject_lora_adapters(model, cfg)
        mark_only_lora_as_trainable(model, cfg)

        # Start: B=0 => LoRA has no effect; merge should be no-op
        with torch.no_grad():
            out0 = model(x)
        merge_lora_weights_(model)
        with torch.no_grad():
            out1 = model(x)
        assert torch.allclose(out0, out1, atol=0, rtol=0)

        # Create a non-zero LoRA: set A random, B random
        for m in model.modules():
            if isinstance(m, LoraInjectedLinear):
                with torch.no_grad():
                    torch.nn.init.normal_(m.lora_B.weight, std=0.02)

        unmerge_lora_weights_(model)  # ensure unmerged mode
        with torch.no_grad():
            out_unmerged = model(x)

        # Merge and check equivalence
        merge_lora_weights_(model)
        with torch.no_grad():
            out_merged = model(x)
        assert torch.allclose(out_unmerged, out_merged, atol=1e-6, rtol=1e-6)

        # Unmerge should restore pre-merge base weights' effect (so outputs equal again)
        unmerge_lora_weights_(model)
        with torch.no_grad():
            out_unmerged2 = model(x)
        assert torch.allclose(out_unmerged, out_unmerged2, atol=1e-6, rtol=1e-6)

    def test_grads_only_on_lora(self):
        x = torch.randn(8, 32)
        y = torch.randn(8, 16)

        model = TinyBlock()
        cfg = LoraConfig(
            r=4, lora_alpha=8, lora_dropout=0.0, target_modules=["q_proj", "v_proj"]
        )
        model, _ = inject_lora_adapters(model, cfg)
        mark_only_lora_as_trainable(model, cfg)

        opt = torch.optim.AdamW(
            [p for p in model.parameters() if p.requires_grad], lr=1e-3
        )
        loss_fn = nn.MSELoss()

        out = model(x)
        loss = loss_fn(out, y)
        loss.backward()

        # No grads on base
        for n, p in model.named_parameters():
            if "lora_" not in n:
                assert p.grad is None, f"Base param {n} unexpectedly has grad"
            else:
                print(f"Base param {n} has grad: {p.grad}")
                assert p.grad is not None, f"LoRA param {n} missing grad"

        # Optim step should change only LoRA params
        old_base = {
            n: p.clone().detach()
            for n, p in model.named_parameters()
            if "lora_" not in n
        }
        old_lora = {
            n: p.clone().detach() for n, p in model.named_parameters() if "lora_" in n
        }

        opt.step()

        with torch.no_grad():
            # Base stays identical
            for n, p in model.named_parameters():
                if "lora_" not in n:
                    assert torch.equal(
                        p, old_base[n]
                    ), f"Base param {n} changed unexpectedly"

            # LoRA must change (use exact inequality, not allclose)
            for n, p in model.named_parameters():
                if "lora_" in n:
                    assert not torch.equal(
                        p, old_lora[n]
                    ), f"LoRA param {n} did not change"

    def test_save_and_load_lora_only(self):
        model = TinyBlock()
        cfg = LoraConfig(
            r=4, lora_alpha=8, lora_dropout=0.0, target_modules=["q_proj", "v_proj"]
        )
        model, _ = inject_lora_adapters(model, cfg)
        mark_only_lora_as_trainable(model, cfg)

        # Make non-zero LoRA
        for m in model.modules():
            if isinstance(m, LoraInjectedLinear):
                with torch.no_grad():
                    torch.nn.init.normal_(m.lora_B.weight, std=0.05)

        # Save only LoRA tensors
        state = lora_state_dict(model, prefix="lora")

        # === Important: align base weights ===
        model2 = TinyBlock()
        model2.load_state_dict(_base_state_dict(model), strict=False)

        # Inject LoRA on model2, then load the LoRA-only checkpoint
        inject_lora_adapters(model2, cfg)
        load_lora_state_dict(model2, state, prefix="lora", strict=True)

        # LoRA params should match exactly
        for (n1, m1), (n2, m2) in zip(model.named_modules(), model2.named_modules()):
            if isinstance(m1, LoraInjectedLinear):
                assert isinstance(m2, LoraInjectedLinear)
                assert torch.allclose(m1.lora_A.weight, m2.lora_A.weight)
                assert torch.allclose(m1.lora_B.weight, m2.lora_B.weight)

        # And with the same base, forward should match
        x = torch.randn(5, 32)
        with torch.no_grad():
            y1 = model(x)
            y2 = model2(x)
        assert torch.allclose(y1, y2, atol=1e-6, rtol=1e-6)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
from torch.distributed.checkpoint.default_planner import DefaultLoadPlanner
from torch.distributed.checkpoint.metadata import STATE_DICT_TYPE, Metadata

from cosmos_rl.utils.logging import logger


class RenameLoadPlanner(DefaultLoadPlanner):
    """
    RenameLoadPlanner that renames variables during checkpoint load.
    """

    def set_up_planner(
        self,
        state_dict: STATE_DICT_TYPE,
        metadata: Optional[Metadata] = None,
        is_coordinator: bool = False,
    ) -> None:
        super().set_up_planner(
            state_dict=state_dict,
            metadata=metadata,
            is_coordinator=is_coordinator,
        )

        self.state_dict = remap_model_state_for_deepep(self.state_dict, metadata)

        # Do an early check to see if the checkpoint is valid and print the missing
        # keys in the state dict if not. The reason is the original default planner's
        # error message is not helpful enough when the keys are mismatched.
        missing_keys = get_missing_keys(self.state_dict, metadata)

        if missing_keys:
            logger.info(f"Missing keys in checkpoint: {missing_keys}...")
            logger.info(f"Checkpoint keys: {list(metadata.state_dict_metadata)}...")


def get_missing_keys(
    state_dict: dict[str, torch.Tensor],
    metadata: Metadata,
) -> list[str]:
    missing_keys = []
    for fqn, obj in state_dict.items():
        # ignore state_dict keys which do not exist in `state_dict` if strict=False
        if fqn not in metadata.state_dict_metadata:
            missing_keys.append(fqn)
    return missing_keys


def remap_model_state_for_deepep(
    state_dict: dict[str, torch.Tensor],
    metadata: Metadata,
) -> dict[str, torch.Tensor]:
    """
    Remap the state dict by removing the "gate_and_up_projs" key.
    And add the "gate_projs" and "up_projs" keys to the state dict.
    """
    import re

    missing_keys = get_missing_keys(state_dict, metadata)

    # Check if there is substring "gate_and_up_projs" in any key of missing_keys
    # If yes, do a remapping of state_dict keys
    needs_remapping = any(["gate_and_up_projs" in key for key in missing_keys])
    if not needs_remapping:
        return state_dict

    logger.info("Old checkpoint, requires remapping of gate_and_up_projs")

    new_state_dict = state_dict.copy()
    for key, v in state_dict.items():
        moe_pattern = r"^([\w.]*)model\.layers\.(\d+)\.mlp\.experts\.gate_and_up_projs$"
        match = re.match(moe_pattern, key)
        if match:
            prefix = match.group(1)
            layer_num = match.group(2)
            logger.info(
                f"Remapping {key} to {prefix}model.layers.{layer_num}.mlp.experts.xxx_projs"
            )

            v_1, v_2 = torch.chunk(v, 2, -1)
            new_state_dict[
                f"{prefix}model.layers.{layer_num}.mlp.experts.gate_projs"
            ] = v_1.transpose(1, 2)
            new_state_dict[f"{prefix}model.layers.{layer_num}.mlp.experts.up_projs"] = (
                v_2.transpose(1, 2)
            )
            del new_state_dict[key]
        elif "mlp.experts.down_projs" in key:
            new_state_dict[key] = v.transpose(1, 2)

    return new_state_dict



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn as nn
from typing import Optional


class RMSNorm(nn.Module):
    def __init__(self, hidden_size, eps=1e-6, casting_mode: Optional[str] = None):
        """
        RMSNorm is equivalent to T5LayerNorm
        casting_mode:
            - "bf16": cast to bf16
            - "fp16": cast to fp16
            - "fp32": cast to fp32
            - None: no casting
        """
        super().__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size))
        self.variance_epsilon = eps
        self.casting_mode = casting_mode

    def forward(self, hidden_states):
        input_dtype = hidden_states.dtype
        hidden_states = hidden_states.to(torch.float32)
        variance = hidden_states.pow(2).mean(-1, keepdim=True)
        hidden_states = hidden_states * torch.rsqrt(variance + self.variance_epsilon)
        return self.weight * hidden_states.to(input_dtype)

    def extra_repr(self):
        return f"{tuple(self.weight.shape)}, eps={self.variance_epsilon}"

    def liger_equivalent(self):
        from liger_kernel.transformers.rms_norm import (
            LigerRMSNorm,
            LigerRMSNormForGemma,
            LigerRMSNormForGemma2,
            LigerRMSNormForGemma3,
        )

        casting_mode_to_liger_norm = {
            "llama": LigerRMSNorm,
            "deepseek_v3": LigerRMSNorm,
            "qwen": LigerRMSNorm,
            "qwen2": LigerRMSNorm,
            "qwen2_5_vl": LigerRMSNorm,
            "qwen3": LigerRMSNorm,
            "qwen3_moe": LigerRMSNorm,
            "gemma": LigerRMSNormForGemma,
            "gemma_2": LigerRMSNormForGemma2,
            "gemma_3": LigerRMSNormForGemma3,
        }
        assert self.casting_mode in casting_mode_to_liger_norm.keys()
        return casting_mode_to_liger_norm[self.casting_mode](
            hidden_size=self.weight.shape[0],
            eps=self.variance_epsilon,
        )
